from typing import Iterator, List, Optional

from abstra_internals.controllers.sdk.sdk_context import SDKContextStore
from abstra_internals.controllers.sdk.sdk_tasks import Task
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.tasks import TaskPayload
from abstra_internals.utils.json import to_json_serializable

BATCH_SIZE = 10


def get_tasks(
    limit: Optional[int] = None, offset: int = 0, where: Optional[dict] = None
) -> List[Task]:
    """Retrieve a list of pending tasks.

    Args:
        limit (Optional[int]): Maximum number of tasks to retrieve.
            Defaults to None which is interpreted as 100.
            Will be capped at 10000.
        offset (int): Number of tasks to skip. Defaults to 0.
        where (Optional[dict]): Dictionary of filters to apply. Defaults to None.

    Returns:
        List[Task]: List of pending tasks matching the criteria.
    """
    context = SDKContextStore.get_by_thread()
    if where is None:
        where = {}

    if limit is None or limit < 1:
        limit = 100
    else:
        if limit > 10000:
            limit = 10000
            AbstraLogger.warning(
                "get_tasks limit is too high, capping it to 10000. To iterate over all tasks, use iter_tasks"
            )

    return context.task_sdk.get_stage_pending_tasks(limit, offset, where)


def send_task(type: str, payload: "TaskPayload", show_warning: bool = True) -> None:
    """Send a new task to the task queue.

    Args:
        type (str): Type identifier for the task.
        payload (TaskPayload): Dictionary containing the task data.
        show_warning (bool): Whether to show warnings. Defaults to True.

    Raises:
        Exception: If the payload is not a dictionary.
    """
    context = SDKContextStore.get_by_thread()
    if not isinstance(payload, dict):
        raise Exception("Task payload must be a dictionary")
    serializable_payload = {k: to_json_serializable(v) for k, v in payload.items()}
    context.task_sdk.send_task(type, serializable_payload, show_warning)


def get_trigger_task() -> Task:
    """Retrieve the task that triggered the current execution.

    Returns:
        Task: The task object that triggered this execution.
    """
    context = SDKContextStore.get_by_thread()
    return context.task_sdk.get_trigger_task()


def iter_tasks(where: Optional[dict] = None) -> Iterator[Task]:
    """Iterate through all pending tasks, handling pagination automatically.

    This function yields tasks preloads in batches of 10. It continues to fetch
    and yield tasks until there are no more tasks to retrieve.

    Args:
        where (Optional[dict]): Dictionary of filters to apply. Defaults to None.

    Yields:
        Task: Each pending task matching the criteria.
    """
    if where is None:
        where = {}
    offset = 0
    while True:
        tasks = get_tasks(limit=BATCH_SIZE, offset=offset, where=where)
        if not tasks:
            break
        yield from tasks
        offset += BATCH_SIZE


def get_sent_tasks(
    limit: Optional[int] = None, offset: int = 0, where: Optional[dict] = None
) -> List[Task]:
    """Retrieve a list of tasks that have been sent.

    Args:
        limit (Optional[int]): Maximum number of tasks to retrieve.
            Defaults to None which is interpreted as 100 if where is also None.
        offset (int): Number of tasks to skip. Defaults to 0.
        where (Optional[dict]): Dictionary of filters to apply. Defaults to None.

    Returns:
        List[Task]: List of sent tasks matching the criteria.
    """
    context = SDKContextStore.get_by_thread()
    if limit is None and where is None:
        limit = 100
    if where is None:
        where = {}
    return context.task_sdk.get_stage_sent_tasks(limit, offset, where)



from unittest import TestCase

from abstra_internals.fs_watcher import FileChangeEventHandler
from abstra_internals.modules import import_as_new
from abstra_internals.repositories.project.project import (
    LocalProjectRepository,
    ScriptStage,
)
from tests.fixtures import clear_dir, init_dir


class ReloadProjectLocalModulesTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()
        self.project_repository = LocalProjectRepository()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_reload(self):
        project = self.project_repository.load()
        mod_code = "x = 1"
        mod_file = self.root / "my_mod.py"
        mod_file.write_text(mod_code)

        main_code = "from my_mod import x"
        main_file = self.root / "main.py"
        main_file.write_text(main_code)
        script = ScriptStage(
            file="main.py",
            id="test",
            is_initial=True,
            title="test",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.scripts.append(script)
        self.project_repository.save(project)

        main_module = import_as_new("main.py")
        self.assertEqual(main_module.x, 1)

        mod_code = "x = 2"
        mod_file.write_text(mod_code)
        handler = FileChangeEventHandler(self.project_repository)
        handler.reload_module(mod_file)
        main_module = import_as_new("main.py")
        self.assertEqual(main_module.x, 2)



from concurrent.futures import ThreadPoolExecutor
from typing import Optional
from uuid import uuid4

from abstra_internals.controllers.execution.execution import ExecutionController
from abstra_internals.controllers.main import MainController
from abstra_internals.entities.execution import ClientContext
from abstra_internals.environment import (
    EXECUTION_QUEUE_CONCURRENCY,
    set_SERVER_UUID,
    set_WORKER_UUID,
)
from abstra_internals.logger import AbstraLogger, Environment
from abstra_internals.repositories.consumer import Consumer, QueueMessage
from abstra_internals.repositories.project.project import StageWithFile
from abstra_internals.settings import Settings
from abstra_internals.stdio_patcher import StdioPatcher


class StageNotFound(Exception):
    pass


class NonCleanExit(Exception):
    pass


# Should be smaller than the RabbitMQ consumer timeout
# to prevent its termination
PROCESS_TIMEOUT_SECONDS = 60 * 60 * 3


class ConsumerController:
    def __init__(self, main_controller: MainController, consumer: Consumer):
        self.main_controller = main_controller
        self.consumer = consumer
        self.app_id = str(uuid4())

    def tagged_msg(self, msg: str) -> str:
        short_id = self.app_id.split("-")[0]
        return f"[ConsumerController ({short_id})] {msg}"

    def start_loop(self):
        AbstraLogger.debug(
            self.tagged_msg(f"Starting loop with {EXECUTION_QUEUE_CONCURRENCY} threads")
        )

        thread_pool = ThreadPoolExecutor(max_workers=EXECUTION_QUEUE_CONCURRENCY)

        for msg in self.consumer.iter():
            try:
                thread_pool.submit(
                    self.run_subprocess,
                    msg=msg,
                )
            except Exception as e:
                AbstraLogger.error(self.tagged_msg(f"iteration error: {e}"))
                AbstraLogger.capture_exception(e)

        AbstraLogger.debug(
            self.tagged_msg("Consumer main loop exited, waiting for threads to finish")
        )
        thread_pool.shutdown(wait=True)
        AbstraLogger.debug(self.tagged_msg("All threads finished"))

        self.main_controller.app_exit(
            app_id=self.app_id,
            err_msg=self.tagged_msg("Consumer main loop exited"),
        )

        raise Exception("Consumer main loop exited")

    def run_subprocess(self, msg: QueueMessage):
        worker_id = str(uuid4())
        head_id = worker_id.split("-")[0]

        try:
            mp_context = self.main_controller.repositories.mp_context.get_context()
            stage = self.main_controller.get_stage(msg.preexecution.stage_id)
            if stage is None:
                raise StageNotFound(msg.preexecution.stage_id)

            p = mp_context.Process(
                target=subprocess,
                kwargs=dict(
                    stage=stage,
                    controller=self.main_controller,
                    worker_id=worker_id,
                    app_id=self.app_id,
                    root_path=Settings.root_path,
                    server_port=Settings.server_port,
                    request=msg.preexecution.context,
                    environment=AbstraLogger.environment,
                ),
                name=f"Worker-{head_id}",
            )

            p.start()
            p.join(timeout=PROCESS_TIMEOUT_SECONDS)

            # If timeout is reached, the process is still alive
            if p.is_alive():
                p.terminate()
                raise NonCleanExit(
                    f"Worker took too long to complete ({PROCESS_TIMEOUT_SECONDS} seconds)"
                )

            if p.exitcode != 0:
                err_msg = f"Worker exited with status ({p.exitcode})"
                if p.exitcode == -9:
                    err_msg += ": Server reached its memory limit"

                raise NonCleanExit(err_msg)

            self.consumer.done_callback(msg)

        except Exception as e:
            AbstraLogger.error(self.tagged_msg(f"Worker {worker_id} error: {e}"))
            AbstraLogger.capture_exception(e)
            self.main_controller.worker_exit(
                app_id=self.app_id,
                worker_id=worker_id,
                err_msg=self.tagged_msg(f"Worker {worker_id} error: {e}"),
            )


def subprocess(
    *,
    root_path: str,
    server_port: int,
    worker_id: str,
    app_id: str,
    stage: StageWithFile,
    controller: MainController,
    environment: Optional[Environment],
    request: Optional[ClientContext] = None,
):
    AbstraLogger.debug("[ConsumerController Subprocess] Starting...")

    Settings.set_root_path(root_path)
    Settings.set_server_port(server_port, force=True)
    AbstraLogger.init(environment)

    set_SERVER_UUID(app_id)
    set_WORKER_UUID(worker_id)

    StdioPatcher.apply(controller)

    ExecutionController(
        repositories=controller.repositories,
        stage=stage,
        client=None,
        context=request,
    ).run()

    AbstraLogger.debug("[ConsumerController Subprocess] Finished")



from abstra_internals.interface.sdk.forms.deprecated.widgets.widget_base import Output


class TextOutput(Output):
    type = "text-output"

    def __init__(self, text: str, **kwargs):
        self.set_props(dict(text=text, **kwargs))

    def set_props(self, props):
        self.text = str(props["text"])
        self.size = props.get("size", "medium")
        self.full_width = props.get("full_width", False)
        super().set_props(props)

    def render(self, ctx: dict):
        return {
            "type": self.type,
            "text": self.text,
            "size": self.size,
            "fullWidth": self.full_width,
        }



import re

from unidecode import unidecode


def normalize_path(path: str):
    paths = filter(_remove_empty_string_filter, path.split("/"))
    return "/".join(map(_normalize_string, paths))


def _remove_empty_string_filter(string: str):
    return string != ""


def _normalize_string(string: str):
    lowercase = string.lower().strip()
    unaccented = unidecode(lowercase)

    alphanumeric_pattern = r"[a-z0-9]+"

    matches = re.findall(alphanumeric_pattern, unaccented)

    return "-".join(matches)


def pip_name(name: str) -> str:
    return name.strip().lower().replace("_", "-")



import time
from threading import Thread
from typing import List, Optional

from abstra_internals.controllers.common.task_executors import TaskExecutor
from abstra_internals.controllers.service.roles.common import RoleCommonController
from abstra_internals.entities.execution import Execution
from abstra_internals.settings import Settings


class RoleClientController(RoleCommonController):
    def entrypoints(self, project_id: str) -> List[dict]:
        return [
            e.to_camel() for e in self.repos.role_clients.get_entrypoints(project_id)
        ]

    def create_foreign_task(
        self,
        task_type: str,
        task_payload: dict,
        target_stage_id: str,
        execution: Optional[Execution] = None,
    ):
        executor = TaskExecutor(self.repos)
        project = self.repos.project.load()
        executor.send_task(
            type=task_type,
            current_stage=project.get_stage_raises(target_stage_id),
            payload=task_payload,
            execution=execution,
        )

    def sync_connection_pool(self):
        self.repos.role_clients.sync_connections()

    def safe_sync_connection_pool(self):
        try:
            self.sync_connection_pool()
        except Exception as e:
            print(e)

    def loop_sync_connection_pool(self):
        def loop():
            while True:
                if Settings.has_public_url():
                    try:
                        self.sync_connection_pool()
                        time.sleep(60 * 5)
                    except Exception as e:
                        print(e)
                        time.sleep(10)  # Delay retry to avoid spamming server
                        pass
                else:
                    time.sleep(10)

        Thread(target=loop, daemon=True).start()



from io import BufferedReader, TextIOWrapper
from pathlib import Path
from typing import List, Optional, Union

from abstra_internals.interface.sdk.forms.deprecated.widgets.response_types import (
    FileResponse,
)
from abstra_internals.interface.sdk.forms.deprecated.widgets.widget_base import (
    Input,
    MultipleHandler,
)
from abstra_internals.utils.file import extension_to_mime, upload_file


class FileInput(Input):
    type = "file-input"
    multiple: bool = False
    accepted_formats: list = []
    multiple_handler: MultipleHandler

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get("label", "Label")
        self.required = props.get("required", True)
        self.hint = props.get("hint", None)
        self.full_width = props.get("full_width", False)
        self.disabled = props.get("disabled", False)
        self.max_file_size = props.get("max_file_size", None)
        self.multiple = props.get("multiple", False)
        self.min = props.get("min", None)
        self.max = props.get("max", None)
        self.accepted_formats = props.get("accepted_formats", [])
        self.value = props.get("initial_value", self.empty_value)
        self.multiple_handler = MultipleHandler(
            self.multiple, self.min, self.max, required=self.required
        )
        super().set_props(props)

    def render(self, ctx: dict):
        return {
            "type": self.type,
            "key": self.key,
            "hint": self.hint,
            "label": self.label,
            "value": self.serialize_value(),
            "required": self.required,
            "multiple": self.multiple,
            "min": self.min,
            "max": self.max,
            "acceptedFormats": self.accepted_formats,
            "acceptedMimeTypes": self.get_mime_types,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "maxFileSize": self.max_file_size,
            "errors": self.errors,
        }

    @staticmethod
    def __get_file_uri(
        value: Union[FileResponse, str, BufferedReader, TextIOWrapper],
    ) -> str:
        if isinstance(value, str):
            if Path(value).is_file():
                return upload_file(open(value, "rb"))
            else:
                return value
        if isinstance(value, FileResponse):
            return value._url
        if isinstance(value, BufferedReader):
            return upload_file(value)
        if isinstance(value, TextIOWrapper):
            return upload_file(value)
        return ""

    def serialize_value(self) -> List[str]:
        values_list = self.multiple_handler.value_to_list(self.value)
        return [FileInput.__get_file_uri(item) for item in values_list]

    def parse_value(
        self, value: Optional[List[str]]
    ) -> Union[FileResponse, List[FileResponse], None]:
        file_responses = [FileResponse(url) for url in value or []]
        return self.multiple_handler.value_to_list_or_value(file_responses)

    def validate(self) -> List[str]:
        errors = []
        errors.extend(self.multiple_handler.validate(self.value))
        if not self.value:
            return errors
        if isinstance(self.value, list):
            for file in self.value:
                errors.extend(self._validate_file_extension(file))
        else:
            errors.extend(self._validate_file_extension(self.value))
        return errors

    @property
    def accepted_formats_without_dot(self) -> List[str]:
        return [extension.replace(".", "") for extension in self.accepted_formats]

    def _validate_file_extension(self, file: FileResponse) -> List[str]:
        if not self.accepted_formats:
            return []
        file_extension = file._url.split(".")[-1]
        if file_extension not in self.accepted_formats_without_dot:
            return ["i18n_error_invalid_file_format"]
        return []

    @property
    def get_mime_types(self) -> str:
        if not self.accepted_formats:
            return "*"
        return ",".join(extension_to_mime(ext) for ext in self.accepted_formats)



from typing import TypedDict, Union

import requests

from abstra_internals.entities.agents import ConnectionModel
from abstra_internals.environment import PROJECT_ID, REQUEST_TIMEOUT, SIDECAR_HEADERS
from abstra_internals.repositories.project.project import Project
from abstra_internals.repositories.services.roles.common import RoleCommonRepository


class FailedResponse(TypedDict):
    error: str
    status_code: int


class RoleAgentRepository(RoleCommonRepository):
    def fetch_connections(self):
        headers = self.get_headers()
        assert headers is not None, "You should be logged in to get connections"
        url = f"{self.base_url}/connections"
        response = requests.get(url, headers=headers, timeout=REQUEST_TIMEOUT)
        if not response.ok:
            raise Exception(response.text)
        self._cached_connections = [
            ConnectionModel.from_camel(connection) for connection in response.json()
        ]
        return self._cached_connections

    def get_connections(self):
        return self.fetch_connections()

    def get_connection_by_stage_id(self, stage_id: str):
        connections = self.get_connections()
        for connection in connections:
            if connection.client_stage_id == stage_id:
                return connection
        raise Exception(f"Connection not found for stage {stage_id}")

    def get_connection_by_token(self, token: str):
        connections = self.get_connections()
        for connection in connections:
            if connection.token == token:
                return connection
        raise Exception(f"Connection not found for token {token}")

    @property
    def role(self):
        return "agent"

    def create_connection_on_cloud(
        self,
        client_proof: str,
        client_stage_id: str,
        agent_stage_id: str,
        client_tasks_url: str,
        project: Project,
    ) -> ConnectionModel:
        """
        Second step of the protocol. As the server, it asks the cloud to build the connection.
        """
        agent_project_id = self.get_current_project_id()
        headers = self.get_headers()
        assert headers is not None, "You should be logged in to connect an agent"
        url = f"{self.base_url}/connections"
        body = {
            "clientStageId": client_stage_id,
            "agentProjectId": agent_project_id,
            "agentStageId": agent_stage_id,
            "clientProof": client_proof,
            "clientTaskUrl": client_tasks_url,
            "agentEntrypoints": [
                ae.to_camel() for ae in project.get_agent_entrypoints()
            ],
        }
        response = requests.post(
            url, headers=headers, json=body, timeout=REQUEST_TIMEOUT
        )
        if not response.ok:
            raise Exception(response.text)
        self.fetch_connections()  # Refresh the cache
        return ConnectionModel.from_camel(response.json())

    def update_connection_on_cloud(
        self, connection_token: str, client_tasks_url: str
    ) -> Union[ConnectionModel, FailedResponse]:
        """
        Used to update the callback url
        """
        headers = self.get_headers()
        assert headers is not None, "You should be logged in to connect an agent"
        url = f"{self.base_url}/connections"
        body = {
            "token": connection_token,
            "clientTaskUrl": client_tasks_url,
        }
        response = requests.patch(
            url, headers=headers, json=body, timeout=REQUEST_TIMEOUT
        )

        if not response.ok:
            error_message = (
                response.json().get("error")
                if "error" in response.json()
                else response.text
            )

            return {
                "status_code": response.status_code,
                "error": error_message,
            }

        self.fetch_connections()  # Refresh the cache
        return ConnectionModel.from_camel(response.json())

    def delete_connection_on_cloud(self, connection_token: str):
        headers = self.get_headers()
        assert headers is not None, "You should be logged in to connect an agent"
        url = f"{self.base_url}/connections"
        body = {
            "tokens": [connection_token],
        }
        response = requests.delete(
            url, headers=headers, json=body, timeout=REQUEST_TIMEOUT
        )
        response.raise_for_status()
        self.fetch_connections()


class ProductionRoleAgentRepository(RoleAgentRepository):
    def get_headers(self):
        return SIDECAR_HEADERS

    def get_current_project_id(self):
        return PROJECT_ID


class LocalRoleAgentRepository(RoleAgentRepository):
    def get_headers(self):
        # TODO: Implement local headers
        raise NotImplementedError

    def get_current_project_id(self):
        # TODO: Implement local project id
        raise NotImplementedError



import copy
from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    TypedDict,
    Union,
)

from abstra_internals.entities.forms.form_state import State
from abstra_internals.entities.forms.widgets.widget_base import InputWidget, Widget


@dataclass
class Button:
    """
    Base class for form navigation buttons.

    Args:
        label (str): The text displayed on the button.
    """

    label: str

    def __init__(self, label: str):
        """Initialize a Button with the given label.

        Args:
            label (str): The text displayed on the button.
        """
        self.label = label


class NextButton(Button):
    """
    Button used to navigate to the next step in a form.
    Automatically labeled with internationalized "Next" text.
    """

    def __init__(self):
        """Initialize a NextButton with the default internationalized label."""
        super().__init__("i18n_next_action")


class BackButton(Button):
    """
    Button used to navigate to the previous step in a form.
    Automatically labeled with internationalized "Back" text.
    """

    def __init__(self):
        """Initialize a BackButton with the default internationalized label."""
        super().__init__("i18n_back_action")


Template = List[Widget]
TemplateWithButtons = Tuple[Template, Optional[List[Button]]]

TemplateFunction = Callable[[State], Optional[Union[Template, TemplateWithButtons]]]
TemplateGenerator = Generator[Union[Template, TemplateWithButtons], None, None]
TemplateGeneratorFunction = Callable[[State], TemplateGenerator]


class RenderOutput(TypedDict):
    widgets: List[Dict]
    has_errors: bool


class TemplateRenderer:
    def __init__(self, template: Template) -> None:
        self._raw_template = template

    @property
    def template(self) -> Template:
        # Avoids using mutable objects as widgets currently have a internal state
        return copy.deepcopy(self._raw_template)

    def make_partial_state(self, raw_state: Dict) -> State:
        parsed = {}
        for idx, widget in enumerate(self.template):
            if (
                isinstance(widget, InputWidget)
                and widget._safe_get_key(idx) in raw_state
            ):
                parsed[widget._safe_get_key(idx)] = widget._parse_value(
                    raw_state.get(widget._safe_get_key(idx))
                )
        return State(parsed)

    def make_state(self, raw_state: Dict) -> State:
        parsed = {}
        for idx, widget in enumerate(self.template):
            if isinstance(widget, InputWidget):
                parsed[widget._safe_get_key(idx)] = widget._parse_value(
                    raw_state.get(widget._safe_get_key(idx))
                )
        return State(parsed)

    def render(self, state: State) -> RenderOutput:
        rendered_widgets = []
        has_errors = False
        for idx, widget in enumerate(self.template):
            if isinstance(widget, InputWidget):
                value = state.get(widget._safe_get_key(idx))
                if value is not None:
                    widget.value = value
                if widget._safe_get_key(idx) in state:
                    widget._apply_validation()
                if len(widget.errors) > 0:
                    has_errors = True

            rendered = widget._render()
            rendered_widgets.append(rendered)

        return RenderOutput(widgets=rendered_widgets, has_errors=has_errors)



from pathlib import Path
from unittest import TestCase

from .paths import NotSubpathError, get_relative_path


class TestRelativePath(TestCase):
    def test_relative_path(self):
        path = Path("foo/bar")
        base_path = Path("foo")
        relative = get_relative_path(path, base_path)
        self.assertEqual(relative, Path("bar"))

    def test_relative_non_related_paths(self):
        path = Path("foo")
        base_path = Path("bar")
        with self.assertRaises(NotSubpathError):
            get_relative_path(path, base_path)

    def test_relative_path_absolute_base(self):
        path = Path("foo")
        base_path = Path("/bar")
        relative = get_relative_path(path, base_path)
        self.assertEqual(relative, Path("foo"))

    def test_both_absolute_paths(self):
        path = Path("/foo/bar")
        base_path = Path("/foo")
        relative = get_relative_path(path, base_path)
        self.assertEqual(relative, Path("bar"))

    def test_both_absolute_non_related_paths(self):
        path = Path("/foo")
        base_path = Path("/bar")
        with self.assertRaises(NotSubpathError):
            get_relative_path(path, base_path)



import json
from pathlib import Path

import requests

from abstra_internals.contracts_generated import CommonTablesSnapshot
from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import CLOUD_API_CLI_URL, REQUEST_TIMEOUT
from abstra_internals.interface.cli.tables.format import tables_snapshot_from_csv
from abstra_internals.interface.cli.tables_messages import (
    bad_json_error_message,
    dump_message,
    file_not_found_error_message,
    generic_error,
    restore_message,
    zod_error_message,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.settings import Settings
from abstra_internals.utils.file import ABSTRA_TABLES_FILE


def dump():
    headers = resolve_headers()
    url = f"{CLOUD_API_CLI_URL}/tables/dump"
    try:
        res = requests.get(url, headers=headers, timeout=REQUEST_TIMEOUT)
        res.raise_for_status()
        res_json = res.json()
        json_str = json.dumps(res_json, indent=2)
        file_path = Settings.root_path.joinpath(ABSTRA_TABLES_FILE)
        file_path.write_text(json_str, encoding="utf-8")
        num_tables = len(res_json["tables"])
        dump_message(num_tables, file_path)
    except Exception as e:
        AbstraLogger.capture_exception(e)
        generic_error()


def restore(dry_run: bool = False, file=ABSTRA_TABLES_FILE):
    headers = resolve_headers()
    url = f"{CLOUD_API_CLI_URL}/tables/restore"
    # /path/to/file
    if Path(file).is_absolute():
        file_path = Path(file).resolve()
    # ~/path/file
    elif file.startswith("~"):
        file_path = Path(file).expanduser().resolve()
    # path/file
    else:
        file_path = Settings.root_path.joinpath(file).resolve()
    try:
        if not file_path.exists():
            raise FileNotFoundError(f"File '{file}' not found.")
        if not file_path.is_file():
            raise FileNotFoundError(f"'{file}' is not a file.")
        if file_path.suffix == ".json":
            json_content = file_path.read_text(encoding="utf-8")
            tables_to_restore = json.loads(json_content)
            parsed_data = CommonTablesSnapshot.from_dict(tables_to_restore)
        elif file_path.suffix == ".csv":
            parsed_data = tables_snapshot_from_csv(file_path)
        else:
            raise ValueError(f"Unsupported file format: {file_path.suffix}")
        if dry_run:
            for table in parsed_data.tables:
                print(f"Table: {table.name}")
                for column in table.columns:
                    print(f"  Column: {column.name}")
                    print(f"    Type: {column.type}")
                    print(f"    Nullable: {column.nullable}")
                    print(f"    Default: {column.default}")
                    print(f"    Foreign Key: {column.foreign_key}")
        else:
            print(parsed_data.to_dict())
            res = requests.post(
                url,
                json=parsed_data.to_dict(),
                headers=headers,
                timeout=REQUEST_TIMEOUT,
            )
            if res.status_code == 400:
                print(res.text)
                zod_error_message(file_path)
                return
            res.raise_for_status()
            num_tables = len(parsed_data.tables)
            restore_message(num_tables)
    except FileNotFoundError:
        file_not_found_error_message(file_path)
    except json.JSONDecodeError:
        bad_json_error_message(file_path)
    except Exception as e:
        AbstraLogger.capture_exception(e)
        generic_error()



from typing import Dict


class State(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        return self.get(key, None)

    def update_values(self, other: Dict):
        for key, value in other.items():
            self[key] = value



from typing import List

from abstra_internals.linter.linter import LinterFix, LinterIssue, LinterRule
from abstra_internals.services.requirements import RequirementsRepository
from abstra_internals.utils import packages as pkg_utils


class AddAbstraToRequirements(LinterFix):
    label = "Add abstra to requirements.txt"

    def fix(self):
        requirements = RequirementsRepository.load()
        local_version = pkg_utils.get_local_package_version()

        requirements.ensure("abstra", str(local_version))
        RequirementsRepository.save(requirements)


class SetAbstraVersionInRequirements(LinterFix):
    label = "Set abstra version in requirements.txt"

    def fix(self):
        requirements = RequirementsRepository.load()
        local_version = pkg_utils.get_local_package_version()

        requirements.ensure("abstra", str(local_version))
        RequirementsRepository.save(requirements)


class AbstraNotInRequirementsFound(LinterIssue):
    def __init__(self) -> None:
        self.label = "Abstra is not in your requirements.txt file."
        self.fixes = [AddAbstraToRequirements()]


class AbstraVersionNotDefined(LinterIssue):
    def __init__(self) -> None:
        self.label = (
            "You have abstra in requirements.txt, but the version is not defined."
        )
        self.fixes = [SetAbstraVersionInRequirements()]


class AbstraVersionInRequirementsIsAheadOfInstalled(LinterIssue):
    def __init__(self) -> None:
        self.label = "The version of abstra in your requirements.txt is ahead of the installed version. Close this server and run 'pip install -r requirements.txt' to fix this."
        self.fixes = []


class AbstraVersionInRequirementsIsBehindInstalled(LinterIssue):
    def __init__(self) -> None:
        self.label = "The version of abstra in your requirements.txt is behind the installed version. Update your requirements.txt to fix this."
        self.fixes = [SetAbstraVersionInRequirements()]


class MissingAbstraInRequirements(LinterRule):
    label = "Abstra should be in your requirements.txt"
    type = "bug"

    def find_issues(self) -> List[LinterIssue]:
        requirements = RequirementsRepository.load()
        local_version = pkg_utils.get_local_package_version()

        if not requirements.has("abstra"):
            return [AbstraNotInRequirementsFound()]

        if requirements.has("abstra", str(local_version)):
            return []

        requirements_version = requirements.get("abstra")
        if requirements_version is None:
            return [AbstraVersionNotDefined()]

        if pkg_utils.parse_version(requirements_version) > local_version:
            return [AbstraVersionInRequirementsIsAheadOfInstalled()]
        else:
            return [AbstraVersionInRequirementsIsBehindInstalled()]



import flask

from abstra_internals.controllers.main import MainController
from abstra_internals.controllers.tasks import DataRequest, TasksController
from abstra_internals.usage import editor_usage


def get_editor_bp(main_controller: MainController) -> flask.Blueprint:
    controller = TasksController(repositories=main_controller.repositories)
    bp = flask.Blueprint("tasks", __name__)

    @bp.post("/list")
    def _get_all_tasks():
        if flask.request.json is None:
            flask.abort(400)
        req = DataRequest.from_dict(flask.request.json)
        tasks, total = controller.list_tasks(req)
        return {
            "tasks": [task.dump() for task in tasks],
            "totalCount": total,
        }

    @bp.get("/<stage_id>")
    @editor_usage
    def _get_tasks(stage_id: str):
        tasks = controller.get_stage_tasks(stage_id)
        return {
            "tasks": [task.dump() for task in tasks],
            "totalCount": len(tasks),
        }

    @bp.get("/<stage_id>/sent")
    def _get_sent_tasks(stage_id: str):
        tasks = controller.get_sent_tasks(stage_id)
        return {
            "tasks": [task.dump() for task in tasks],
            "totalCount": len(tasks),
        }

    @bp.put("/<task_id>/status")
    @editor_usage
    def _update_task_status(task_id: str):
        if not flask.request.json:
            flask.abort(400)
        status = flask.request.json.get("status")
        controller.update_task_status(task_id, status)
        return {"success": True}

    @bp.post("/")
    @editor_usage
    def _create_task():
        if not flask.request.json:
            flask.abort(400)
        stage_id = flask.request.json.get("stage_id")
        payload = flask.request.json.get("payload")
        name = flask.request.json.get("name")
        task = controller.create_task(name, stage_id, payload)
        return task.dump()

    @bp.delete("/clear")
    @editor_usage
    def _clear_tasks():
        controller.clear_tasks()
        return {"success": True}

    return bp



from abstra_internals.entities.forms.steps import (
    EndPageStep,
    TemplateFunction,
)


def end_page_step(func: "TemplateFunction") -> EndPageStep:
    """
    Add this decorator to a function to mark it as the end page of a form.
    It will remove buttons and end execution.

    Args:
        func (TemplateFunction): The function to be decorated.

    Returns:
        EndPageStep: A step object that represents the end page of a form.
    """
    return EndPageStep(func)



import json
from dataclasses import dataclass
from typing import Any, Dict, List

from abstra_internals.cloud_api import (
    cancel_all,
    create_thread,
    generate_project,
    get_ai_messages,
    get_history,
)
from abstra_internals.controllers.linters import fix_all_linters
from abstra_internals.controllers.main import MainController
from abstra_internals.credentials import resolve_headers
from abstra_internals.repositories.project.json_migrations import get_latest_version
from abstra_internals.repositories.project.project import (
    Project,
    StageWithFile,
)
from abstra_internals.services.env_vars import EnvVarsRepository
from abstra_internals.settings import Settings
from abstra_internals.utils.file import silent_traverse_code
from abstra_internals.utils.paths import get_relative_path


@dataclass
class PythonFile:
    filename: str
    content: str
    stage: str

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "PythonFile":
        return PythonFile(data["filename"], data["content"], data["stage"])


def runtime_from_stage(stage: StageWithFile):
    if stage.type_name == "form":
        return "forms"
    elif stage.type_name == "hook":
        return "hooks"
    elif stage.type_name == "script":
        return "scripts"
    elif stage.type_name == "job":
        return "jobs"
    else:
        raise ValueError(f"Unknown stage type: {stage.type_name}")


def find_imported_code(stage: StageWithFile) -> Dict[str, str]:
    seen_files = set()
    conflict_counter = 0
    files = {}
    for path in silent_traverse_code(stage.file_path):
        if path in seen_files:
            conflict_counter += 1
            if conflict_counter > 10:  # Circular imports?
                break
            continue
        seen_files.add(path)
        if stage.file_path.absolute() == path.absolute():
            continue
        relative_path = str(get_relative_path(path, Settings.root_path))
        files[relative_path] = "\n".join(
            [
                f"# {relative_path}",
                "```python",
                path.read_text(encoding="utf-8"),
                "````",
            ]
        )
    return files


class AiController:
    def __init__(self, controller: MainController):
        self.controller = controller

    def send_ai_message(
        self, messages, stage_id, langgraph_thread_id, code, execution_error
    ):
        headers = resolve_headers() or {}
        env_vars_keys = EnvVarsRepository.list_keys()
        current_abstra_json = self.controller.repositories.project.load().as_dict
        runtime = None
        imported_code = {}
        if stage_id:
            stage = self.controller.get_stage(stage_id)
            assert isinstance(stage, StageWithFile), "Stage is not a StageWithFile"
            imported_code = find_imported_code(stage)
            code = "\n".join([f"# {stage.file_path}", "```python", code, "````"])
            runtime = runtime_from_stage(stage)
        yield from get_ai_messages(
            messages,
            runtime,
            langgraph_thread_id,
            code,
            imported_code,
            execution_error,
            headers,
            env_vars_keys,
            current_abstra_json,
        )

    def get_history(self, limit: int, offset: int):
        headers = resolve_headers()
        if headers is None:
            return None
        return get_history(headers, limit, offset)

    def create_thread(self):
        headers = resolve_headers()
        if headers is None:
            return None
        return create_thread(headers)

    def cancel_all(self, thread_id: str):
        if headers := resolve_headers():
            cancel_all(headers, thread_id)

    def generate_project(self, prompt: str, tries: int = 0):
        headers = resolve_headers() or {}
        abstra_json_version = get_latest_version()
        try:
            res = generate_project(prompt, abstra_json_version, headers)
            is_abstra_json_valid = Project.validate(res["abstra_json"])
            if not is_abstra_json_valid:
                raise Exception("Generated abstra.json is not valid")

            generated_abstra_json = json.dumps(res["abstra_json"], indent=4)
            python_files = [PythonFile.from_dict(file) for file in res["python_files"]]

            self.init_stages(python_files)
            Settings.root_path.joinpath("abstra.json").write_text(
                generated_abstra_json, encoding="utf-8"
            )

            self.controller.repositories.project.initialize_or_migrate(verbose=False)
            fix_all_linters()
        except Exception as e:
            if tries >= 3:
                raise e
            return self.generate_project(prompt, tries + 1)

    def init_stages(self, python_files: List[PythonFile]):
        for file in python_files:
            if file.stage == "form":
                script = self.controller.create_form(file.filename[:-3], file.filename)
            elif file.stage == "hook":
                script = self.controller.create_hook(file.filename[:-3], file.filename)
            elif file.stage == "script":
                script = self.controller.create_script(
                    file.filename[:-3], file.filename
                )
            elif file.stage == "job":
                script = self.controller.create_job(file.filename[:-3], file.filename)
            else:
                raise Exception(f"Invalid stage {file.stage}")
            script.file_path.write_text(file.content, encoding="utf-8")



from abstra_internals.controllers.main import MainController


class AgentsController:
    url: str

    def __init__(self, controller: MainController):
        self.controller = controller

    def get_agent_connections(self):
        return self.controller.repositories.agents.get_agent_connections()

    def get_is_usage_mode(self):
        return self.controller.repositories.agents.get_is_usage_mode()



from .FileInput import FileInput


class CameraInput(FileInput):
    """Camera input widget for capturing images or videos directly from a camera.

    Attributes:
        value (Union[FileResponse, List[FileResponse], None]): The captured image or video file(s).
    """

    type = "camera-input"



import abc
from typing import TypedDict, Union

import requests

from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import REQUEST_TIMEOUT, SIDECAR_HEADERS


class TokenValidationResult(TypedDict):
    is_valid: bool
    client_project_id: Union[str, None]
    error: Union[str, None]


class AgentsRepository(abc.ABC):
    url: str

    def __init__(self, base_url: str) -> None:
        self.url = f"{base_url}/agents"

    def validate_token(self, token: str) -> TokenValidationResult:
        if not token:
            raise ValueError("Token is required")

        response = requests.post(
            f"{self.url}/validate_token",
            headers=self.get_headers(),
            json={
                "token": token,
            },
            timeout=REQUEST_TIMEOUT,
        )
        if not response.ok:
            return {
                "is_valid": False,
                "client_project_id": None,
                "error": response.text,
            }

        returned_data = response.json()

        return {
            "is_valid": returned_data.get("isValid"),
            "client_project_id": returned_data.get("clientProjectId"),
            "error": returned_data.get("error"),
        }

    def accept_connection(self, token: str):
        if not token:
            raise ValueError("Token is required")

        response = requests.post(
            f"{self.url}/accept_connection",
            headers=self.get_headers(),
            json={
                "token": token,
            },
            timeout=REQUEST_TIMEOUT,
        )
        response.raise_for_status()

        return response.json()

    def get_agent_connections(self):
        url = f"{self.url}/connections"
        response = requests.get(
            url, headers=self.get_headers(), timeout=REQUEST_TIMEOUT
        )
        response.raise_for_status()
        return response.json()

    def get_is_usage_mode(self):
        try:
            url = f"{self.url}/is_usage_mode"
            response = requests.get(
                url, headers=self.get_headers(), timeout=REQUEST_TIMEOUT
            )
            response.raise_for_status()
            return {"is_usage_mode": response.json()}
        except Exception:
            return {"is_usage_mode": False}

    def get_headers(self) -> dict:
        raise NotImplementedError()


class LocalAgentsRepository(AgentsRepository):
    def get_headers(self) -> dict:
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to work with agents")
        return headers


class ProductionAgentsRepository(AgentsRepository):
    def get_headers(self) -> dict:
        return SIDECAR_HEADERS



from .base_migration import Migration


class Migration013(Migration):
    @staticmethod
    def target_version() -> str:
        return "13.0"

    @staticmethod
    def source_version() -> str:
        return "12.0"

    def _migrate(self) -> None:
        self.data["agents"] = []
        self.data["clients"] = []



import typing
from typing import Optional, Tuple

import requests

from abstra_internals.email_templates import authn_template
from abstra_internals.environment import OIDC_AUTHORITY, REQUEST_TIMEOUT
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.email import EmailRepository
from abstra_internals.repositories.jwt_signer import JWTRepository
from abstra_internals.repositories.keyvalue import KVRepository
from abstra_internals.repositories.users import UsersRepository


def authn_kv_key(email: str) -> str:
    return f"authn::key::{email}"


def authn_kv_counter(email: str) -> str:
    return f"authn::counter::{email}"


def get_oidc_userinfo_email(access_token: str) -> typing.Optional[str]:
    if not OIDC_AUTHORITY():
        raise ValueError("[OIDC] OIDC_AUTHORITY is not set")

    r = requests.get(
        f"{OIDC_AUTHORITY()}/.well-known/openid-configuration",
        timeout=REQUEST_TIMEOUT,
    )
    if not r.ok:
        raise ValueError("[OIDC] Failed to fetch OIDC configuration")

    userinfo_endpoint = r.json().get("userinfo_endpoint")
    if not userinfo_endpoint:
        raise ValueError("[OIDC] userinfo_endpoint not found in OIDC configuration")

    r = requests.get(
        userinfo_endpoint,
        headers={"Authorization": f"Bearer {access_token}"},
        timeout=REQUEST_TIMEOUT,
    )
    if not r.ok:
        return None

    return r.json().get("email")


class AuthController:
    kv_repository: KVRepository
    jwt_repository: JWTRepository
    users_repository: UsersRepository
    email_repository: EmailRepository

    def __init__(
        self,
        *,
        kv_repository: KVRepository,
        jwt_repository: JWTRepository,
        email_repository: EmailRepository,
        users_repository: UsersRepository,
    ):
        self.kv_repository = kv_repository
        self.jwt_repository = jwt_repository
        self.email_repository = email_repository
        self.users_repository = users_repository

    def _sign_and_signup(self, email: str) -> str:
        self.users_repository.insert_user(email)
        return self.jwt_repository.sign(email)

    def authenticate(self, email: str) -> bool:
        try:
            key = authn_kv_key(email)
            code = self.jwt_repository.gen_code()
            self.kv_repository.set(key, code)

            counter_key = authn_kv_counter(email)
            self.kv_repository.set(counter_key, ".")

            authentication_email = authn_template.generate_email(email, code)
            self.email_repository.send(authentication_email)

            return True
        except Exception as e:
            AbstraLogger.capture_exception(e)
            return False

    def verify(self, email: str, user_code: str) -> Tuple[int, Optional[str]]:
        try:
            key = authn_kv_key(email)
            counter_key = authn_kv_counter(email)
            user_code = self.jwt_repository.sanitize_code(user_code)

            code = self.kv_repository.get(key)
            if code != user_code:
                counter = self.kv_repository.get(counter_key)
                if len(counter) >= 3:
                    self.kv_repository.delete(key)
                    self.kv_repository.delete(counter_key)
                    return 410, None

                self.kv_repository.set(counter_key, counter + ".")
                return 400, None

            self.kv_repository.delete(key)
            self.kv_repository.delete(counter_key)

            return 200, self._sign_and_signup(email)
        except Exception as e:
            AbstraLogger.capture_exception(e)
            return 500, None

    def oidc_verify(self, access_token: str) -> Tuple[Optional[str], Optional[str]]:
        try:
            email = get_oidc_userinfo_email(access_token)
            if not email:
                return None, None

            return self._sign_and_signup(email), email
        except Exception as e:
            AbstraLogger.capture_exception(e)
            return None, None



from abc import ABC, abstractmethod
from typing import List

import requests

from abstra_internals.contracts_generated import CommonRole
from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import REQUEST_TIMEOUT


class RolesRepository(ABC):
    @abstractmethod
    def get_roles(self) -> List[CommonRole]:
        raise NotImplementedError


class LocalRolesRepository(RolesRepository):
    def __init__(self, url: str) -> None:
        self.url = url

    @property
    def headers(self):
        return resolve_headers()

    def get_roles(self) -> List[CommonRole]:
        if not self.headers:
            return []

        response = requests.get(
            f"{self.url}/roles", headers=self.headers, timeout=REQUEST_TIMEOUT
        )
        return [CommonRole.from_dict(role) for role in response.json()]


class ProductionRolesRepository(RolesRepository):
    def __init__(self, url: str) -> None:
        self.url = url

    def get_roles(self) -> List[CommonRole]:
        # not applicable for production
        raise NotImplementedError()



from .roles import (
    LocalRoleAgentRepository,
    LocalRoleClientRepository,
    ProductionRoleAgentRepository,
    ProductionRoleClientRepository,
)

__all__ = [
    "ProductionRoleClientRepository",
    "ProductionRoleAgentRepository",
    "LocalRoleClientRepository",
    "LocalRoleAgentRepository",
]



import jedi
from pydantic import BaseModel


def _script(code: str):
    return jedi.Script(code, environment=jedi.InterpreterEnvironment())


def jedi_get_autocomplete(code, line, column):
    try:
        script = _script(code)
        completions = script.complete(line, column)
        result = []
        for c in completions:
            completion = {
                "params": [],
                "name": c.name,
                "source": "jedi",
                "documentation": c.docstring(),
            }
            for completion_signatures in c.get_signatures():
                for p in completion_signatures.params:
                    completion["params"].append({"name": p.name})
            result.append(completion)

        return result
    except Exception:
        return []


def jedi_help(code, line, column):
    try:
        script = _script(code)
        return [
            {"docstring": h.docstring(), "name": h.name}
            for h in script.help(line, column)
        ]
    except Exception:
        return []


def jedi_get_syntax_errors(code):
    try:
        script = _script(code)
        errors = script.get_syntax_errors()
        return [
            {
                "line": e.line,
                "column": e.column,
                "until_line": e.until_line,
                "until_column": e.until_column,
                "message": e.get_message(),
                "severity": "error",
            }
            for e in errors
        ]
    except Exception:
        return []


class Context(BaseModel):
    code: str
    line: int = 0
    column: int = 0



from pathlib import Path
from pkgutil import iter_modules

from abstra_internals.linter.linter import LinterFix, LinterIssue, LinterRule
from abstra_internals.repositories.project.project import LocalProjectRepository
from abstra_internals.settings import Settings


class AddPreffix(LinterFix):
    label = "Fix conflicting name"
    preffix = "util_"

    def __init__(self, file: Path):
        self.file = file
        self.project_repository = LocalProjectRepository()

    def fix(self):
        new_file = self.file.parent / f"{self.preffix}{self.file.name}"

        self.file.rename(new_file)

        project = self.project_repository.load()

        for stage in project.get_stages_by_file_path(self.file):
            project.update_stage(stage, dict(file=str(new_file.name)))

        self.project_repository.save(project)


class ConflictingNameIssue(LinterIssue):
    type = "bug"
    fixes = []

    def __init__(self, file: Path):
        self.label = f"The name of the file {file.name} is in conflict with an internal reserved name. This can cause unexpected behavior. You can either change it manually in the Editor or use the 'Fix conflicting name' button."
        self.fixes = [AddPreffix(file)]


def reserved_names():
    return [
        m.name
        for m in iter_modules()
        if m not in list(iter_modules([str(Settings.root_path)]))
    ]


class ConflictingName(LinterRule):
    label = "Conflicting path"
    type = "bug"

    def find_issues(self):
        root = Settings.root_path
        project_py_files = set(root.glob("*.py"))
        _reserved_names = set(root / (n + ".py") for n in reserved_names())

        issues = [
            ConflictingNameIssue(file)
            for file in _reserved_names.intersection(project_py_files)
        ]

        return issues



from typing import List, Optional

from pydantic import BaseModel
from pydantic.alias_generators import to_camel, to_snake


class AgentTaskSchema(BaseModel):
    task_type: str
    task_payload_schema: Optional[dict]


class AgentEntrypoint(BaseModel):
    title: str
    client_stage_id: str
    input_schemas: Optional[List[AgentTaskSchema]]
    output_schemas: Optional[List[AgentTaskSchema]]

    @staticmethod
    def from_camel(data: dict):
        snake_data = {to_snake(key): value for key, value in data.items()}
        return AgentEntrypoint.model_validate(snake_data)

    def to_camel(self):
        return {to_camel(key): value for key, value in self.model_dump().items()}


class ConnectionModel(BaseModel):
    token: str
    client_project_id: str
    client_stage_id: str
    client_task_url: str
    agent_project_id: str
    agent_stage_id: str

    @staticmethod
    def from_camel(data: dict):
        snake_data = {to_snake(key): value for key, value in data.items()}
        return ConnectionModel.model_validate(snake_data)


class AgentModel(BaseModel):
    project_id: str
    tasks_url: str


class SignProofResponse(BaseModel):
    signedProof: str



import json
import os
import shutil
import sys
import tempfile
import uuid
from abc import ABC, abstractmethod
from dataclasses import field
from pathlib import Path
from threading import Lock
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple, Union

from pydantic.dataclasses import dataclass

from abstra_internals.constants import ABSTRA_LOGO_URL, get_project_url
from abstra_internals.contracts_generated import (
    CommonAbstraJsonV13,
    CommonAbstraJsonV13AgentsItem,
    CommonAbstraJsonV13ClientsItem,
    CommonAbstraJsonV13DefinitionsFormStage,
    CommonAbstraJsonV13DefinitionsFormStageAccessControl,
    CommonAbstraJsonV13DefinitionsFormStageNotificationTrigger,
    CommonAbstraJsonV13DefinitionsHookStage,
    CommonAbstraJsonV13DefinitionsJobStage,
    CommonAbstraJsonV13DefinitionsScriptStage,
    CommonAbstraJsonV13DefinitionsTransition,
    CommonAbstraJsonV13Home,
    CommonAbstraJsonV13HomeAccessControl,
    CommonAbstraJsonV13Workspace,
    CommonUser,
)
from abstra_internals.entities.agents import AgentEntrypoint
from abstra_internals.environment import IS_PRODUCTION
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.project import json_migrations
from abstra_internals.settings import Settings
from abstra_internals.templates import (
    abstra_favicon,
    abstra_logo,
)
from abstra_internals.utils import check_is_url, nested_get
from abstra_internals.utils.file import generate_conflictless_path, silent_traverse_code
from abstra_internals.utils.format import normalize_path
from abstra_internals.utils.graph import Edge, Graph, Node
from abstra_internals.utils.string import to_kebab_case

ServedStage = Union["FormStage", "HookStage"]
StageType = Literal["form", "hook", "job", "script", "agent", "client"]


class Stage(ABC):
    id: str
    title: str
    workflow_transitions: List["WorkflowTransition"]
    workflow_position: Tuple[float, float]
    type_name: StageType

    @property
    @abstractmethod
    def as_dict(self) -> dict:
        pass

    @property
    @abstractmethod
    def editor_dto(self) -> dict:
        pass

    @property
    @abstractmethod
    def admin_dto(self) -> dict:
        pass

    @abstractmethod
    def update(self, changes: Dict[str, Any]):
        pass

    @abstractmethod
    def duplicate(self, new_id: str, new_position: Tuple[int, int] = (0, 0)) -> "Stage":
        pass

    def __post_init__(self):
        self.workflow_transitions = self.workflow_transitions or []


class StageWithFile(Stage):
    file: str

    @property
    @abstractmethod
    def file_path(self) -> Path:
        pass


SecuredStage = Union["FormStage", "Home"]

Languages = Literal[
    "en",
    "pt",
    "de",
    "es",
    "fr",
    "hi",
]


@dataclass
class NotificationTrigger:
    variable_name: str
    enabled: bool

    def validate_email(self, email: str) -> bool:
        return isinstance(email, str) and "@" in email

    def get_recipients(self, task_payload: Dict[str, Any]) -> List[str]:
        variable_name = self.variable_name
        if not variable_name:
            return []

        raw_value = nested_get(task_payload, variable_name)
        if not raw_value:
            return []

        emails: List[str] = []
        if isinstance(raw_value, str):
            emails.append(raw_value)

        if isinstance(raw_value, list):
            emails = raw_value

        return [email for email in emails if self.validate_email(email)]

    def to_dto(self) -> Optional[Dict[str, Any]]:
        return {"variable_name": self.variable_name, "enabled": self.enabled}

    def to_abstra_json_dto(
        self,
    ) -> CommonAbstraJsonV13DefinitionsFormStageNotificationTrigger:
        return CommonAbstraJsonV13DefinitionsFormStageNotificationTrigger(
            variable_name=self.variable_name, enabled=self.enabled
        )


@dataclass
class WorkflowTransition:
    id: str
    target_id: str
    target_type: str
    type: str
    task_type: Optional[str] = None

    def __post_init__(self):
        self.id = str(uuid.uuid4()) if self.id is None else self.id

    def matches(self, task_type: str):
        if self.task_type is None or self.task_type.strip() == "":
            return True
        elif isinstance(task_type, str):
            return self.task_type == task_type

    @property
    def as_dict(self) -> dict:
        return {
            "target_id": self.target_id,
            "target_type": self.target_type,
            "type": self.type,
            "id": self.id,
            "task_type": self.task_type,
        }

    @staticmethod
    def from_dict(data: dict):
        if "type" not in data:
            raise Exception("Transition type is required")
        return WorkflowTransition(
            target_id=data["target_id"],
            target_type=data["target_type"],
            type=data["type"],
            id=data["id"],
            task_type=data.get("task_type"),
        )

    @staticmethod
    def default_type(source_type: StageType) -> str:
        if source_type == "form":
            return "forms:finished"
        elif source_type == "hook":
            return "hooks:finished"
        elif source_type == "job":
            return "jobs:finished"
        elif source_type == "script":
            return "scripts:finished"
        else:
            raise Exception(f"Invalid source type {source_type}")

    def to_abstra_json_dto(self) -> CommonAbstraJsonV13DefinitionsTransition:
        return CommonAbstraJsonV13DefinitionsTransition(
            id=self.id,
            target_id=self.target_id,
            target_type=self.target_type,
            type=self.type,
            task_type=self.task_type,
        )


@dataclass
class HookStage(StageWithFile):
    id: str
    file: str
    path: str
    title: str
    workflow_transitions: List[WorkflowTransition]
    workflow_position: Tuple[float, float]
    is_initial: bool = True
    enabled: bool = False
    type_name = "hook"

    @staticmethod
    def from_dict(dto: dict):
        x, y = dto["workflow_position"]
        return HookStage(
            id=dto["id"],
            file=dto["file"],
            path=dto["path"],
            title=dto["title"],
            enabled=dto["enabled"],
            workflow_position=(x, y),
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in dto["transitions"]
            ],
            is_initial=dto["is_initial"],
        )

    @staticmethod
    def create(
        title: str,
        file: str,
        id: Union[str, None] = None,
        workflow_position: Tuple[int, int] = (0, 0),
    ):
        _id: str = id or str(uuid.uuid4())
        return HookStage(
            id=_id,
            file=file,
            path=to_kebab_case(file[:-3]),
            title=title,
            is_initial=True,
            workflow_transitions=[],
            workflow_position=workflow_position,
        )

    @property
    def admin_dto(self):
        return {
            "title": self.title,
            "id": self.id,
            "path": self.path,
            "type": "hook",
            "is_initial": self.is_initial,
        }

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "file": self.file,
            "path": self.path,
            "title": self.title,
            "enabled": self.enabled,
            "workflow_position": self.workflow_position,
            "is_initial": self.is_initial,
            "transitions": [t.as_dict for t in self.workflow_transitions],
        }

    @property
    def editor_dto(self):
        return self.as_dict

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in ["title", "enabled"]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            elif attr == "path":
                setattr(self, attr, normalize_path(value))
            else:
                raise Exception(f"Cannot update {attr} of hook")

    def duplicate(self, new_id: str, new_position: Tuple[int, int] = (0, 0)):
        return self.from_dict(
            {
                **self.as_dict,
                "id": new_id,
                "workflow_position": new_position,
                "transitions": [],
            }
        )

    def to_abstra_json_dto(self) -> CommonAbstraJsonV13DefinitionsHookStage:
        return CommonAbstraJsonV13DefinitionsHookStage(
            id=self.id,
            file=self.file,
            path=self.path,
            title=self.title,
            enabled=self.enabled,
            workflow_position=[self.workflow_position[0], self.workflow_position[1]],
            transitions=[t.to_abstra_json_dto() for t in self.workflow_transitions],
            is_initial=self.is_initial,
        )


@dataclass
class ScriptStage(StageWithFile):
    id: str
    file: str
    title: str
    workflow_transitions: List[WorkflowTransition]
    workflow_position: Tuple[float, float]
    is_initial: bool = True
    type_name = "script"

    @staticmethod
    def create(
        title: str,
        file: str,
        id: Union[str, None] = None,
        workflow_position: Tuple[int, int] = (0, 0),
    ):
        _id = id or str(uuid.uuid4())
        return ScriptStage(
            id=_id,
            file=file,
            title=title,
            is_initial=True,
            workflow_transitions=[],
            workflow_position=workflow_position,
        )

    @staticmethod
    def from_dict(data: Dict):
        x, y = data["workflow_position"]
        return ScriptStage(
            id=data["id"],
            file=data["file"],
            title=data["title"],
            workflow_position=(x, y),
            is_initial=data["is_initial"],
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in data["transitions"]
            ],
        )

    @property
    def admin_dto(self):
        return {
            "title": self.title,
            "id": self.id,
            "type": "script",
            "is_initial": self.is_initial,
        }

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "file": self.file,
            "title": self.title,
            "is_initial": self.is_initial,
            "workflow_position": self.workflow_position,
            "transitions": [t.as_dict for t in self.workflow_transitions],
        }

    @property
    def editor_dto(self):
        return self.as_dict

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in ["id", "title", "schedule"]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            else:
                raise Exception(f"Cannot update {attr} of script")

    def duplicate(self, new_id: str, new_position: Tuple[int, int] = (0, 0)):
        return self.from_dict(
            {
                **self.as_dict,
                "id": new_id,
                "workflow_position": new_position,
                "transitions": [],
            }
        )

    def to_abstra_json_dto(self) -> CommonAbstraJsonV13DefinitionsScriptStage:
        return CommonAbstraJsonV13DefinitionsScriptStage(
            id=self.id,
            file=self.file,
            title=self.title,
            workflow_position=[self.workflow_position[0], self.workflow_position[1]],
            is_initial=self.is_initial,
            transitions=[t.to_abstra_json_dto() for t in self.workflow_transitions],
        )


@dataclass
class AccessSettings:
    is_public: bool
    required_roles: List[str]

    def should_allow(self, user: Optional[CommonUser]) -> bool:
        if self.is_public:
            return True
        if not user:
            return False
        return len(self.required_roles) == 0 or any(
            role in user.roles for role in self.required_roles
        )

    @staticmethod
    def from_dict(data: dict):
        return AccessSettings(
            is_public=data.get("is_public", True),
            required_roles=data.get("required_roles", []),
        )

    @property
    def as_dict(self):
        return {
            "is_public": self.is_public,
            "required_roles": self.required_roles,
        }

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            setattr(self, attr, value)

    def to_abstra_json_dto(
        self,
    ) -> CommonAbstraJsonV13DefinitionsFormStageAccessControl:
        return CommonAbstraJsonV13DefinitionsFormStageAccessControl(
            is_public=self.is_public, required_roles=self.required_roles
        )


@dataclass
class JobStage(StageWithFile):
    id: str
    file: str
    title: str
    schedule: str
    workflow_position: Tuple[float, float]
    workflow_transitions: List[WorkflowTransition]
    type_name = "job"

    is_initial = True

    @staticmethod
    def create(
        title: str,
        file: str,
        id: Union[str, None] = None,
        workflow_position: Tuple[int, int] = (0, 0),
    ):
        _id = id or str(uuid.uuid4())
        return JobStage(
            id=_id,
            file=file,
            title=title,
            schedule="0 0 * * *",
            workflow_position=workflow_position,
            workflow_transitions=[],
        )

    @staticmethod
    def from_dict(data: Dict):
        x, y = data["workflow_position"]
        return JobStage(
            file=data["file"],
            id=data["id"],
            title=data["title"],
            schedule=data["schedule"],
            workflow_position=(x, y),
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in data["transitions"]
            ],
        )

    @property
    def admin_dto(self):
        return {"title": self.title, "id": self.id, "type": "job", "is_initial": True}

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "file": self.file,
            "title": self.title,
            "schedule": self.schedule,
            "workflow_position": self.workflow_position,
            "transitions": [t.as_dict for t in self.workflow_transitions],
        }

    @property
    def editor_dto(self):
        return self.as_dict

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in ["identifier", "title", "schedule"]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            else:
                raise Exception(f"Cannot update {attr} of job")

    def duplicate(self, new_id: str, new_position: Tuple[int, int] = (0, 0)):
        return self.from_dict(
            {
                **self.as_dict,
                "id": new_id,
                "workflow_position": new_position,
                "transitions": [],
            }
        )

    def to_abstra_json_dto(self) -> CommonAbstraJsonV13DefinitionsJobStage:
        return CommonAbstraJsonV13DefinitionsJobStage(
            id=self.id,
            file=self.file,
            title=self.title,
            schedule=self.schedule,
            workflow_position=[self.workflow_position[0], self.workflow_position[1]],
            transitions=[t.to_abstra_json_dto() for t in self.workflow_transitions],
        )


@dataclass
class FormStage(StageWithFile):
    id: str
    file: str
    title: str
    path: str
    workflow_transitions: List[WorkflowTransition]
    workflow_position: Tuple[float, float]
    notification_trigger: NotificationTrigger
    is_initial: bool = True
    end_message: Optional[str] = None
    auto_start: Optional[bool] = False
    start_message: Optional[str] = None
    error_message: Optional[str] = None
    welcome_title: Optional[str] = None
    allow_restart: Optional[bool] = False
    timeout_message: Optional[str] = None
    start_button_text: Optional[str] = None
    restart_button_text: Optional[str] = None
    access_control: AccessSettings = field(
        default_factory=lambda: AccessSettings(is_public=False, required_roles=[])
    )

    type_name = "form"

    @staticmethod
    def create(
        title: str,
        file: str,
        id: Union[str, None] = None,
        workflow_position: Tuple[int, int] = (0, 0),
    ):
        _id = id or str(uuid.uuid4())
        return FormStage(
            id=_id,
            file=file,
            path=generate_conflictless_path(to_kebab_case(file[:-3])),
            title=title,
            is_initial=True,
            workflow_transitions=[],
            workflow_position=workflow_position,
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
            access_control=AccessSettings(is_public=False, required_roles=[]),
        )

    @staticmethod
    def from_dict(data: dict):
        x, y = data["workflow_position"]
        return FormStage(
            id=data["id"],
            file=data["file"],
            path=data["path"],
            title=data["title"],
            end_message=data["end_message"],
            auto_start=data["auto_start"],
            start_message=data["start_message"],
            error_message=data["error_message"],
            welcome_title=data["welcome_title"],
            allow_restart=data["allow_restart"],
            timeout_message=data["timeout_message"],
            start_button_text=data["start_button_text"],
            restart_button_text=data["restart_button_text"],
            workflow_position=(x, y),
            is_initial=data["is_initial"],
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in data["transitions"]
            ],
            notification_trigger=NotificationTrigger(
                variable_name=data["notification_trigger"]["variable_name"],
                enabled=data["notification_trigger"]["enabled"],
            ),
            access_control=AccessSettings.from_dict(data.get("access_control", {})),
        )

    @property
    def admin_dto(self):
        return {
            "title": self.title,
            "id": self.id,
            "path": self.path,
            "type": "form",
        }

    @property
    def to_sidebar_item(self) -> "SidebarItem":
        return SidebarItem(
            id=self.id,
            name=self.title,
            path=self.path,
            type="form",
        )

    @property
    def browser_runner_dto(self):
        allow_restart = self.allow_restart if self.is_initial else False

        return {
            "id": self.id,
            "path": self.path,
            "title": self.title,
            "is_initial": self.is_initial,
            "auto_start": self.auto_start,
            "allow_restart": allow_restart,
            "end_message": self.end_message,
            "start_message": self.start_message,
            "error_message": self.error_message,
            "welcome_title": self.welcome_title,
            "timeout_message": self.timeout_message,
            "start_button_text": self.start_button_text,
            "restart_button_text": self.restart_button_text,
        }

    @property
    def editor_dto(self):
        return {
            **self.browser_runner_dto,
            "file": self.file,
            "workflow_position": self.workflow_position,
            "transitions": [t.as_dict for t in self.workflow_transitions],
            "notification_trigger": (
                self.notification_trigger.to_dto()
                if self.notification_trigger
                else None
            ),
            "access_control": self.access_control.as_dict,
        }

    @property
    def as_dict(self):
        return self.editor_dto

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in [
                "title",
                "end_message",
                "auto_start",
                "start_message",
                "error_message",
                "welcome_title",
                "allow_restart",
                "timeout_message",
                "start_button_text",
                "restart_button_text",
            ]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            elif attr == "access_control":
                self.access_control = AccessSettings.from_dict(value)
            elif attr == "path":
                setattr(self, attr, generate_conflictless_path(normalize_path(value)))
            elif attr == "notification_trigger":
                self.notification_trigger = NotificationTrigger(
                    variable_name=value["variable_name"], enabled=value["enabled"]
                )
            else:
                raise Exception(f"Cannot update {attr} of form")

    def duplicate(self, new_id: str, new_position: Tuple[int, int] = (0, 0)):
        return self.from_dict(
            {
                **self.as_dict,
                "id": new_id,
                "workflow_position": new_position,
                "transitions": [],
            }
        )

    def to_access_dto(self):
        return {
            "id": self.id,
            "title": self.title,
            "type": self.type_name,
            "is_public": self.access_control.is_public,
            "required_roles": self.access_control.required_roles,
        }

    def to_abstra_json_dto(self) -> CommonAbstraJsonV13DefinitionsFormStage:
        return CommonAbstraJsonV13DefinitionsFormStage(
            id=self.id,
            file=self.file,
            path=self.path,
            title=self.title,
            end_message=self.end_message,
            auto_start=self.auto_start or False,
            start_message=self.start_message,
            error_message=self.error_message,
            welcome_title=self.welcome_title,
            allow_restart=self.allow_restart or False,
            timeout_message=self.timeout_message,
            start_button_text=self.start_button_text,
            restart_button_text=self.restart_button_text,
            workflow_position=[self.workflow_position[0], self.workflow_position[1]],
            transitions=[t.to_abstra_json_dto() for t in self.workflow_transitions],
            is_initial=self.is_initial,
            notification_trigger=self.notification_trigger.to_abstra_json_dto(),
            access_control=self.access_control.to_abstra_json_dto(),
        )


@dataclass
class AgentStage(Stage):
    id: str
    title: str
    project_id: Optional[str]
    workflow_transitions: List[WorkflowTransition]
    workflow_position: Tuple[float, float]
    client_stage_id: Optional[str]
    type_name = "agent"

    @staticmethod
    def from_dict(data: Dict):
        x, y = data["workflow_position"]
        return AgentStage(
            id=data["id"],
            title=data["title"],
            project_id=data["project_id"],
            client_stage_id=data["client_stage_id"],
            workflow_position=(x, y),
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in data["transitions"]
            ],
        )

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "title": self.title,
            "project_id": self.project_id,
            "workflow_position": self.workflow_position,
            "transitions": [t.as_dict for t in self.workflow_transitions],
            "client_stage_id": self.client_stage_id,
        }

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in ["title", "project_id"]:
                setattr(self, attr, value)
            else:
                raise Exception(f"Cannot update {attr} of agent")

    @staticmethod
    def create(title: str, project_id: str, id: Union[str, None] = None):
        _id = id or str(uuid.uuid4())
        return AgentStage(
            id=_id,
            title=title,
            project_id=project_id,
            client_stage_id="",
            workflow_transitions=[],
            workflow_position=(0, 0),
        )

    @property
    def editor_dto(self):
        return self.as_dict

    @property
    def admin_dto(self):
        return {
            "title": self.title,
            "id": self.id,
            "type": "agent",
        }

    def duplicate(self, new_id: str, new_position: Tuple[int, int] = (0, 0)):
        return self.from_dict(
            {
                **self.as_dict,
                "id": new_id,
                "workflow_position": new_position,
                "transitions": [],
            }
        )

    def to_abstra_json_dto(self) -> CommonAbstraJsonV13AgentsItem:
        return CommonAbstraJsonV13AgentsItem(
            id=self.id,
            title=self.title,
            project_id=self.project_id,
            client_stage_id=self.client_stage_id,
            workflow_position=[self.workflow_position[0], self.workflow_position[1]],
            transitions=[t.to_abstra_json_dto() for t in self.workflow_transitions],
        )


@dataclass
class ClientStage(Stage):
    id: str
    title: str
    workflow_transitions: List[WorkflowTransition]
    workflow_position: Tuple[float, float]
    type_name = "client"

    @staticmethod
    def from_dict(data: Dict):
        x, y = data["workflow_position"]
        return ClientStage(
            id=data["id"],
            title=data["title"],
            workflow_position=(x, y),
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in data["transitions"]
            ],
        )

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "title": self.title,
            "workflow_position": self.workflow_position,
            "transitions": [t.as_dict for t in self.workflow_transitions],
        }

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in ["id", "title"]:
                setattr(self, attr, value)
            else:
                raise Exception(f"Cannot update {attr} of script")

    @property
    def editor_dto(self):
        return self.as_dict

    @property
    def admin_dto(self):
        return {
            "title": self.title,
            "id": self.id,
            "type": "client",
        }

    def duplicate(self, new_id: str, new_position: Tuple[int, int] = (0, 0)):
        return self.from_dict(
            {
                **self.as_dict,
                "id": new_id,
                "workflow_position": new_position,
                "transitions": [],
            }
        )

    def get_agent_entrypoint(self):
        return AgentEntrypoint(
            title=self.title,
            client_stage_id=self.id,
            input_schemas=None,
            output_schemas=None,
        )

    def to_abstra_json_dto(self) -> CommonAbstraJsonV13ClientsItem:
        return CommonAbstraJsonV13ClientsItem(
            id=self.id,
            title=self.title,
            workflow_position=[self.workflow_position[0], self.workflow_position[1]],
            transitions=[t.to_abstra_json_dto() for t in self.workflow_transitions],
        )


@dataclass
class SidebarItem:
    id: str
    name: str
    path: str
    type: str

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "path": self.path,
            "type": self.type,
        }

    @staticmethod
    def from_dict(data: dict):
        return SidebarItem(
            id=data["id"],
            name=data["name"],
            path=data["path"],
            type=data["type"],
        )


@dataclass
class Sidebar:
    items: List[SidebarItem]

    @staticmethod
    def from_dict(
        sidebar_data: List[dict],
    ):
        return Sidebar(items=[SidebarItem.from_dict(item) for item in sidebar_data])

    @property
    def as_dict(self):
        return [item.as_dict for item in self.items]


@dataclass
class StyleSettings:
    name: str
    language: Languages
    theme: Optional[str]
    logo_url: Optional[str]
    favicon_url: Optional[str]
    brand_name: Optional[str]
    main_color: Optional[str]
    font_family: Optional[str]
    font_color: Optional[str]

    @property
    def project_name(self):
        return self.brand_name or "Abstra Project"

    @property
    def project_logo_url(self):
        if self.logo_url:
            if check_is_url(self.logo_url):
                return self.logo_url

            if IS_PRODUCTION:
                return f"{get_project_url()}/_assets/logo"

        return ABSTRA_LOGO_URL

    @property
    def as_dict(self):
        default_value = StyleSettings.defaultValue()
        return {
            k: v or default_value.__dict__.get(k, v) for k, v in self.__dict__.items()
        }

    @property
    def browser_runner_dto(self):
        if isinstance(self.logo_url, str) and check_is_url(self.logo_url):
            logo_url = self.logo_url
        elif self.logo_url:
            logo_url = "/_assets/logo"
        else:
            logo_url = None

        if isinstance(self.favicon_url, str) and check_is_url(self.favicon_url):
            favicon_url = self.favicon_url
        elif self.favicon_url:
            favicon_url = "/_assets/favicon.ico"
        else:
            favicon_url = None

        return {**self.as_dict, "logo_url": logo_url, "favicon_url": favicon_url}

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr == "sidebar":
                continue
            else:
                setattr(self, attr, value)

    @staticmethod
    def from_dict(data: Dict):
        defaultValue = StyleSettings.defaultValue()
        return StyleSettings(
            name=data.get("name", defaultValue.name),
            font_family=data.get("font_family", defaultValue.font_family),
            brand_name=data.get("brand_name", defaultValue.brand_name),
            main_color=data.get("main_color", defaultValue.main_color),
            font_color=data.get("font_color", defaultValue.font_color),
            logo_url=data.get("logo_url", defaultValue.logo_url),
            favicon_url=data.get("favicon_url", defaultValue.favicon_url),
            theme=data.get("theme", defaultValue.theme),
            language=data.get("language", defaultValue.language),
        )

    @staticmethod
    def create():
        return StyleSettings(
            name="Workspace",
            font_family=None,
            brand_name="Abstra Project",
            main_color=None,
            font_color=None,
            logo_url="./logo.png",
            favicon_url="./favicon.ico",
            theme=None,
            language="en",
        )

    @staticmethod
    def defaultValue():
        return StyleSettings(
            name="Workspace",
            font_family=None,
            brand_name="",
            main_color=None,
            font_color=None,
            logo_url=None,
            favicon_url=None,
            theme=None,
            language="en",
        )

    def to_abstra_json_dto(self) -> CommonAbstraJsonV13Workspace:
        return CommonAbstraJsonV13Workspace(
            name=self.name,
            language=self.language,
            theme=self.theme,
            logo_url=self.logo_url,
            favicon_url=self.favicon_url,
            brand_name=self.brand_name,
            main_color=self.main_color,
            font_family=self.font_family,
            font_color=self.font_color,
        )


@dataclass
class StyleSettingsWithSidebar(StyleSettings):
    sidebar: Sidebar

    @staticmethod
    def from_dict(data: Dict):
        child_data = StyleSettings.from_dict(data)
        return StyleSettingsWithSidebar(
            **{
                **child_data.__dict__,
                "sidebar": Sidebar.from_dict(data.get("sidebar", [])),
            }
        )

    @property
    def as_dict(self):
        return {
            **super().as_dict,
            "sidebar": self.sidebar.as_dict,
        }


class PathConflictError(Exception):
    pass


class StageNotFoundError(KeyError):
    pass


class TransitionNotFoundError(KeyError):
    pass


@dataclass
class Home:
    access_control: AccessSettings

    @staticmethod
    def from_dict(data: dict):
        return Home(
            access_control=AccessSettings.from_dict(
                data.get("access_control", {"is_public": False, "required_roles": []})
            ),
        )

    @staticmethod
    def create():
        return Home(
            access_control=AccessSettings(is_public=False, required_roles=[]),
        )

    @property
    def as_dict(self):
        return {"access_control": self.access_control.as_dict}

    @property
    def type_name(self):
        return "home"

    @property
    def path(self):
        return ""

    @property
    def title(self):
        return "Home"

    @property
    def id(self):
        return "home"

    def to_access_dto(self):
        return {
            "id": self.id,
            "title": self.title,
            "type": self.type_name,
            "is_public": self.access_control.is_public,
            "required_roles": self.access_control.required_roles,
        }

    @property
    def to_sidebar_item(self) -> SidebarItem:
        return SidebarItem(
            id=self.id,
            name=self.title,
            path=self.path,
            type=self.type_name,
        )

    def update(self, id, changes: Dict[str, Any]):
        if id == "access_control":
            setattr(self, id, AccessSettings.from_dict(changes))

    def to_abstra_json_dto(self) -> CommonAbstraJsonV13Home:
        return CommonAbstraJsonV13Home(
            access_control=CommonAbstraJsonV13HomeAccessControl(
                is_public=self.access_control.is_public,
                required_roles=self.access_control.required_roles,
            )
        )


@dataclass
class Project:
    workspace: StyleSettings
    home: Home
    scripts: List[ScriptStage]
    forms: List[FormStage]
    hooks: List[HookStage]
    jobs: List[JobStage]
    agents: List[AgentStage]
    clients: List[ClientStage]

    _graph: Graph

    @property
    def as_dict(self):
        target_stages = set()
        for stage in self.workflow_stages:
            for transition in stage.workflow_transitions:
                target_stages.add(transition.target_id)

        for stage in self.jobs + self.forms + self.scripts + self.hooks:
            if stage.id in target_stages:
                stage.is_initial = False
            else:
                stage.is_initial = True

        return {
            "workspace": self.workspace.as_dict,
            "home": self.home.as_dict,
            "jobs": [job.as_dict for job in self.jobs],
            "hooks": [hook.as_dict for hook in self.hooks],
            "forms": [form.as_dict for form in self.forms],
            "scripts": [script.as_dict for script in self.scripts],
            "agents": [agent.as_dict for agent in self.agents],
            "clients": [client.as_dict for client in self.clients],
        }

    def to_abstra_json_dto(self):
        return CommonAbstraJsonV13(
            version="13.0",
            workspace=self.workspace.to_abstra_json_dto(),
            home=self.home.to_abstra_json_dto(),
            jobs=[job.to_abstra_json_dto() for job in self.jobs],
            hooks=[hook.to_abstra_json_dto() for hook in self.hooks],
            forms=[form.to_abstra_json_dto() for form in self.forms],
            scripts=[script.to_abstra_json_dto() for script in self.scripts],
            agents=[agent.to_abstra_json_dto() for agent in self.agents],
            clients=[client.to_abstra_json_dto() for client in self.clients],
        )

    @property
    def workflow_stages(self) -> List[Stage]:
        return [
            *self.forms,
            *self.jobs,
            *self.hooks,
            *self.scripts,
            *self.agents,
            *self.clients,
        ]

    def get_stages_by_file_path(self, file_path: Path) -> List[StageWithFile]:
        stage_with_file_classes = (FormStage, HookStage, JobStage, ScriptStage)
        return [
            stage
            for stage in self.workflow_stages
            if isinstance(stage, stage_with_file_classes)
            and stage.file_path.resolve() == file_path.resolve()
        ]

    def iter_entrypoints(self) -> Generator[Path, None, None]:
        for stage in self.workflow_stages:
            if isinstance(stage, StageWithFile):
                yield Path(stage.file)

    def iter_entrypointed_stages(self) -> Generator[Tuple[Path, Stage], None, None]:
        for stage in self.workflow_stages:
            if isinstance(stage, StageWithFile):
                yield Path(stage.file), stage

    @property
    def project_files(self) -> Generator[Path, None, None]:
        for entrypoint in self.iter_entrypoints():
            yield from silent_traverse_code(entrypoint)

    def get_local_dependencies(self) -> Generator[Path, None, None]:
        entrypoints = list(self.iter_entrypoints())
        for file in self.project_files:
            if file not in entrypoints:
                yield file

    def get_next_stages_ids(self, id: str) -> List[str]:
        return [node.id for node in self._graph.next_to(id)]

    def get_previous_stages_ids(self, id: str) -> List[str]:
        return [node.id for node in self._graph.previous_to(id)]

    def is_initial(self, target_stage: Stage) -> bool:
        return len(self._graph.previous_to(target_stage.id)) == 0

    def add_stage(self, stage: Stage):
        if isinstance(stage, FormStage):
            self.forms.append(stage)
        elif isinstance(stage, HookStage):
            self.hooks.append(stage)
        elif isinstance(stage, JobStage):
            self.jobs.append(stage)
        elif isinstance(stage, ScriptStage):
            self.scripts.append(stage)
        elif isinstance(stage, AgentStage):
            self.agents.append(stage)
        elif isinstance(stage, ClientStage):
            self.clients.append(stage)
        else:
            raise Exception(f"Cannot add stage of type {type(stage)}")

    def get_workspace(self):
        sidebar = [stage.to_sidebar_item.as_dict for stage in self.secured_stages]
        return StyleSettingsWithSidebar.from_dict(
            {**self.workspace.as_dict, "sidebar": sidebar}
        )

    def default_sidebar(self) -> Sidebar:
        return Sidebar(items=[stage.to_sidebar_item for stage in self.secured_stages])

    def get_access_control_by_stage_id(self, id: str) -> Optional[AccessSettings]:
        for stage in [self.home, *self.forms, *self.jobs]:
            if stage.id == id:
                return stage.access_control
        return None

    def get_access_control_by_stage_path(self, path: str) -> Optional[AccessSettings]:
        for stage in self.secured_stages:
            if stage.path == path:
                return stage.access_control
        return None

    @property
    def secured_stages(self) -> List[SecuredStage]:
        return [self.home, *self.forms]

    def get_secured_stage(self, id: str) -> Optional[SecuredStage]:
        for stage in self.secured_stages:
            if stage.id == id:
                return stage

    def update_access_controls(self, changes: List[Dict[str, Any]]):
        response = []
        for change in changes:
            id = change["id"]
            del change["id"]
            response.append(self.update_access_control(id, change))

        return response

    def update_access_control(self, id: str, change: Dict[str, Any]):
        stage = self.get_secured_stage(id)
        if not stage:
            raise StageNotFoundError(f"Stage with id '{id}' not found")
        if isinstance(stage, Home):
            stage.update("access_control", change)
            return stage.to_access_dto()
        stage = self.update_stage(stage, {"access_control": change})
        if not isinstance(stage, (FormStage)):
            raise Exception(
                f"Stage with id {stage.id} is a {type(stage.type_name)} does not have access control"
            )

        return stage.to_access_dto()

    def get_stage_raises(self, id: str) -> Stage:
        stage = self.get_stage(id)
        if not stage:
            raise StageNotFoundError(f"Stage with id '{id}' not found")
        return stage

    def get_stage(self, id: str) -> Optional[Stage]:
        for stage in self.workflow_stages:
            if stage.id == id:
                return stage

        return None

    def get_form(self, id: str) -> Optional[FormStage]:
        for form in self.forms:
            if form.id == id:
                return form

        return None

    def get_hook(self, id: str) -> Optional[HookStage]:
        for hook in self.hooks:
            if hook.id == id:
                return hook

        return None

    def get_job(self, id: str) -> Optional[JobStage]:
        for job in self.jobs:
            if job.id == id:
                return job

        return None

    def get_script(self, id: str) -> Optional[ScriptStage]:
        for script in self.scripts:
            if script.id == id:
                return script

        return None

    def get_hook_by_path(self, path: str) -> Optional[HookStage]:
        for hook in self.hooks:
            if hook.path == path:
                return hook

        return None

    def get_form_by_path(self, path: str) -> Optional[FormStage]:
        for form in self.forms:
            if form.path == path:
                return form

    def get_stage_by_path(self, path: str) -> Optional[Stage]:
        for form in self.forms:
            if form.path == path:
                return form

        for hook in self.hooks:
            if hook.path == path:
                return hook

        return None

    def get_initial_stages(self) -> List[Stage]:
        return [r for r in self.workflow_stages if self.is_initial(r)]

    def delete_transition_by_target(self, target_id: str):
        for stage in self.workflow_stages:
            stage.workflow_transitions = [
                t for t in stage.workflow_transitions if t.target_id != target_id
            ]

    def get_agent_entrypoints(self) -> List[AgentEntrypoint]:
        return [client.get_agent_entrypoint() for client in self.clients]

    def update_path(self, stage: ServedStage, new_path: str):
        old_path = stage.path

        if self.get_stage_by_path(new_path):
            raise PathConflictError(f"Path {new_path} already exists")

        all_stages: List[Union[FormStage, HookStage, JobStage]] = [
            *self.forms,
            *self.hooks,
            *self.jobs,
        ]

        for _ in all_stages:
            for transition in stage.workflow_transitions:
                if transition.target_id == old_path:
                    transition.target_id = new_path

        stage.path = new_path

    def update_stage(self, stage: Stage, changes: Dict[str, Any]) -> Stage:
        # This guarantees that the updated stage will be the one from the project
        project_stage = self.get_stage(stage.id)

        if project_stage is None:
            raise StageNotFoundError(f"Stage with id '{stage.id}' not found")

        new_path = changes.get("path")

        if new_path:
            if not isinstance(project_stage, FormStage) and not isinstance(
                project_stage, HookStage
            ):
                raise Exception(
                    f"Stage with id {id} is a {type(project_stage)} does not have a path"
                )

            normalized_path = normalize_path(new_path)
            self.update_path(project_stage, normalized_path)

        project_stage.update(changes)

        return project_stage

    def delete_stage(self, id: str, remove_file: bool = False):
        stage = self.get_stage(id)

        if remove_file and isinstance(stage, StageWithFile):
            if not stage:
                return
            path = Settings.root_path.joinpath(stage.file)
            if path.exists():
                os.remove(path.absolute())

        self.delete_transition_by_target(id)
        self.forms = [f for f in self.forms if f.id != id]
        self.hooks = [h for h in self.hooks if h.id != id]
        self.jobs = [j for j in self.jobs if j.id != id]
        self.scripts = [s for s in self.scripts if s.id != id]
        self.agents = [a for a in self.agents if a.id != id]
        self.clients = [c for c in self.clients if c.id != id]

    @staticmethod
    def __from_dict(data: dict):
        target_stages = set()
        nodes = []
        edges = []

        stage_keys = ["forms", "hooks", "scripts", "jobs", "agents", "clients"]

        for key in stage_keys:
            for stage in data[key]:
                nodes.append(Node(id=stage["id"]))
                for transition in stage.get("transitions", []):
                    target_stages.add(transition.get("target_id"))
                    edges.append(
                        Edge(source_id=stage["id"], target_id=transition["target_id"])
                    )

        for key in stage_keys:
            for index, stage in enumerate(data[key]):
                if stage["id"] in target_stages:
                    data[key][index]["is_initial"] = False
                else:
                    data[key][index]["is_initial"] = True

        scripts = [ScriptStage.from_dict(script) for script in data["scripts"]]
        forms = [FormStage.from_dict(form) for form in data["forms"]]
        hooks = [HookStage.from_dict(hook) for hook in data["hooks"]]
        jobs = [JobStage.from_dict(job) for job in data["jobs"]]
        agents = [AgentStage.from_dict(agent) for agent in data["agents"]]
        clients = [ClientStage.from_dict(client) for client in data["clients"]]

        workspace = StyleSettings.from_dict(data["workspace"])
        home = Home.from_dict(data.get("home", {}))

        return Project(
            workspace=workspace,
            scripts=scripts,
            forms=forms,
            hooks=hooks,
            jobs=jobs,
            agents=agents,
            clients=clients,
            home=home,
            _graph=Graph.from_primitives(nodes=nodes, edges=edges),
        )

    @staticmethod
    def validate(data: dict):
        try:
            Project.__from_dict(data)
            return True
        except Exception:
            return False

    @staticmethod
    def from_dict(data: dict):
        try:
            return Project.__from_dict(data)
        except Exception as e:
            print("Error: incompatible abstra.json file.")
            AbstraLogger.capture_exception(e)
            sys.exit(1)

    @staticmethod
    def create():
        return Project(
            workspace=StyleSettings.create(),
            scripts=[],
            forms=[],
            hooks=[],
            jobs=[],
            agents=[],
            clients=[],
            home=Home.create(),
            _graph=Graph.from_primitives([], []),
        )


def _update_file(
    stage: Stage,
    new_file_relative: str,
):
    if not isinstance(stage, StageWithFile):
        return

    old_file = Settings.root_path.joinpath(stage.file)
    new_file = Settings.root_path.joinpath(new_file_relative)

    if old_file.exists() and not new_file.exists():
        new_file.parent.mkdir(parents=True, exist_ok=True)
        old_file.rename(new_file)

    stage.file = new_file_relative


class ProjectRepository:
    lock = Lock()

    def get_file_path(self):
        return Settings.root_path / "abstra.json"

    def initialize(self):
        if not self.exists():  # double check
            self.save(Project.create())
            self.add_assets()

    def add_assets(self):
        logo_path = Settings.root_path / "logo.png"
        if not logo_path.exists():
            logo_path.write_bytes(abstra_logo)

        favicon_path = Settings.root_path / "favicon.ico"
        if not favicon_path.exists():
            favicon_path.write_bytes(abstra_favicon)

    def exists(self):
        return self.get_file_path().exists()

    def save(self, project: Project):
        temp_file = Path(tempfile.mkdtemp()) / "abstra.json"

        project_dto = project.to_abstra_json_dto()

        with temp_file.open("w") as f:
            json.dump(project_dto.to_dict(), f, indent=2)

        with self.lock:
            shutil.move(str(temp_file), self.get_file_path())
            Path.rmdir(temp_file.parent)

    def migrate_config_file(self, verbose=True):
        if not self.exists():
            return
        data = json.loads(self.get_file_path().read_text(encoding="utf-8"))
        assert isinstance(data, dict)
        initial_version = data.get("version")

        migrated_data = json_migrations.migrate(
            data,
            Settings.root_path,
            verbose=verbose,
        )

        if migrated_data["version"] != initial_version:
            self.save(Project.from_dict(migrated_data))

    def load(self) -> Project:
        file_path = self.get_file_path()

        with self.lock:
            data = json.loads(file_path.read_text(encoding="utf-8"))
            if data["version"] == "v13.0":
                CommonAbstraJsonV13.from_dict(data)
            return Project.from_dict(data)

    def initialize_or_migrate(self, verbose=True):
        if not self.exists():
            self.initialize()
        else:
            self.migrate_config_file(verbose=verbose)


class LocalProjectRepository(ProjectRepository):
    pass


class ProductionProjectRepository(ProjectRepository):
    pass



from typing import List, Optional, Tuple

from abstra_internals.controllers.common.task_executors import TaskExecutor
from abstra_internals.controllers.service.roles.common import RoleCommonController
from abstra_internals.entities.execution import Execution
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.project.project import (
    Project,
    Stage,
)
from abstra_internals.repositories.tasks import TaskPayload


class RoleAgentController(RoleCommonController):
    def _get_task_targets(
        self,
        current_stage: Stage,
        type: str,
        task_payload: TaskPayload,
        project: Project,
        show_warning: bool = True,
    ) -> List[Tuple[Stage, dict]]:
        targets = []

        possible_transitions = list(
            filter(
                lambda t: True if not t.task_type else type == t.task_type,
                current_stage.workflow_transitions,
            )
        )

        for transition in possible_transitions:
            target_stage = project.get_stage(transition.target_id)
            if not target_stage:
                continue

            targets.append((target_stage, task_payload))

        if len(targets) == 0 and show_warning:
            AbstraLogger.warning(
                f"Task with payload {task_payload} was deleted, no transitions were satisfied with the type: {type}"
            )

        return targets

    def create_foreign_task(
        self,
        task_type: str,
        task_payload: dict,
        target_stage_id: str,
        execution: Optional[Execution] = None,
    ):
        executor = TaskExecutor(self.repos)
        project = self.repos.project.load()
        executor.send_task(
            type=task_type,
            current_stage=project.get_stage_raises(target_stage_id),
            payload=task_payload,
            execution=execution,
        )



from typing import Optional

import flask

from abstra_internals.contracts_generated import AbstraLibApiEditorWebEditorResponse
from abstra_internals.environment import WAITING_ROOM_URL
from abstra_internals.repositories.jwt_signer import EditorJWTRepository


class WebEditorController:
    def __init__(
        self,
        editor_jwt: EditorJWTRepository,
    ) -> None:
        self.jwt = editor_jwt

    def inspect(self) -> AbstraLibApiEditorWebEditorResponse:
        return AbstraLibApiEditorWebEditorResponse(waiting_room_url=WAITING_ROOM_URL)

    def set_cookie(self, token: Optional[str]):
        if not token:
            return flask.make_response({"ok": False, "error": "No token provided"}, 401)

        if not self.jwt.verify(token):
            return flask.make_response({"ok": False, "error": "Invalid token"}, 401)

        response = flask.make_response({"ok": True})
        response.set_cookie(
            "editor_auth",
            token,
            max_age=60 * 60 * 24 * 7,  # Expires in 7 day (in seconds)
            httponly=True,  # Makes it inaccessible to JavaScript
            secure=True,  # Ensures it's sent over HTTPS
            samesite="Lax",  # Protects against CSRF
        )
        return response

    def guard(self):
        # Local development uses a fake JWT repository
        token = flask.request.cookies.get("editor_auth")
        if not self.jwt.verify(token):
            return flask.make_response({"ok": False, "error": "Invalid token"}, 403)
        return None



import flask
from flask import Blueprint

from abstra_internals.controllers.main import MainController
from abstra_internals.controllers.service.roles.client import RoleClientController


def get_player_bp(main_controller: MainController) -> Blueprint:
    controller = RoleClientController(main_controller.repositories)

    bp = Blueprint("role_client", __name__)

    @bp.post("/")
    def _create_task():
        if not flask.request.json:
            flask.abort(400)
        auth = flask.request.headers.get("authorization")
        if auth is None:
            raise Exception("Authorization header is required")
        payload = flask.request.json
        connection = next(
            c
            for c in controller.repos.role_clients.get_connections()
            if c.token == auth
        )

        controller.create_foreign_task(
            task_type=payload["type"],
            task_payload=payload["payload"],
            target_stage_id=connection.client_stage_id,
        )
        return {"success": True}

    return bp


def get_editor_bp(main_controller: MainController) -> Blueprint:
    bp = Blueprint("role_client_editor", __name__)
    controller = RoleClientController(main_controller.repositories)

    @bp.get("/entrypoints/<string:project_id>")
    def _entrypoints(project_id):
        return controller.entrypoints(project_id)

    return bp



import threading

from dotenv import load_dotenv
from werkzeug.serving import make_server

from abstra_internals.cloud_api import connect_tunnel
from abstra_internals.controllers.execution.consumer import ConsumerController
from abstra_internals.controllers.main import MainController
from abstra_internals.controllers.service.roles.client import RoleClientController
from abstra_internals.environment import HOST
from abstra_internals.fs_watcher import FileChangeEventHandler
from abstra_internals.interface.cli.messages import serve_message
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.consumer import EditorConsumer
from abstra_internals.repositories.factory import get_editor_repositories
from abstra_internals.repositories.producer import LocalProducerRepository
from abstra_internals.resources_watcher import resources_polling_loop
from abstra_internals.server.apps import get_local_app
from abstra_internals.settings import Settings
from abstra_internals.stdio_patcher import StdioPatcher
from abstra_internals.utils.browser import browser_open_editor
from abstra_internals.version import check_latest_version


def start_consumer(controller: MainController):
    if not isinstance(controller.producer_repository, LocalProducerRepository):
        raise ValueError("Invalid producer repository")

    consumer = EditorConsumer(controller.producer_repository.queue)

    th = threading.Thread(
        daemon=True,
        name="execution_consumer",
        target=ConsumerController(controller, consumer).start_loop,
    )

    th.start()

    return consumer, th


def start_file_watcher(watcher: FileChangeEventHandler):
    threading.Thread(
        daemon=True,
        name="file_watcher",
        target=watcher.run,
    ).start()


def start_resources_watcher():
    threading.Thread(
        daemon=True,
        name="resources_watcher",
        target=resources_polling_loop,
    ).start()


def editor(headless: bool):
    load_dotenv(Settings.root_path / ".env")
    serve_message()
    check_latest_version()
    AbstraLogger.init("local")

    controller = MainController(repositories=get_editor_repositories())
    controller.reset_repositories()
    StdioPatcher.apply(controller)

    watcher = FileChangeEventHandler(controller.repositories.project)
    start_file_watcher(watcher)

    start_resources_watcher()
    start_consumer(controller)

    role_client_controller = RoleClientController(controller.repositories)
    role_client_controller.loop_sync_connection_pool()

    app = get_local_app(controller)
    server = make_server(host=HOST, port=Settings.server_port, threaded=True, app=app)

    if not headless:
        browser_open_editor()

    connect_tunnel(
        on_public_url_update=role_client_controller.safe_sync_connection_pool
    )

    server.serve_forever()



import abc
import typing
from dataclasses import dataclass
from typing import Any, Optional

import requests

from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import REQUEST_TIMEOUT, SIDECAR_HEADERS


@dataclass
class ColumnDTO:
    id: str
    name: str
    type: str
    table_id: str
    unique: bool
    default: str
    nullable: bool
    primary_key: bool
    protected: bool
    foreign_key: Optional[str]

    @staticmethod
    def from_dict(data: dict) -> "ColumnDTO":
        return ColumnDTO(
            id=data["id"],
            name=data["name"],
            type=data["type"],
            table_id=data["tableId"],
            unique=data["unique"],
            default=data["default"],
            nullable=data["nullable"],
            primary_key=data["primaryKey"],
            protected=data["protected"],
            foreign_key=data["foreignKey"],
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class TableDTO:
    id: str
    name: str
    project_id: str

    @staticmethod
    def from_dict(data: dict) -> "TableDTO":
        return TableDTO(
            id=data["id"],
            name=data["name"],
            project_id=data["projectId"],
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


class TablesApiHttpClient(abc.ABC):
    base_url: str
    execute_url: str

    def __init__(self, base_url: str) -> None:
        self.execute_url = f"{base_url}/tables/execute"
        self.base_url = base_url

    def execute(self, query: str, params: typing.List) -> requests.Response:
        raise NotImplementedError()

    def create_table(self, id: str, name: str) -> TableDTO:
        raise NotImplementedError()

    def create_column(self, table_id: str, name: str, type: str) -> ColumnDTO:
        raise NotImplementedError()

    def insert_row(self, table_id: str, values: dict):
        raise NotImplementedError()

    def update_table(self, table_id: str, name: str) -> TableDTO:
        raise NotImplementedError()

    def update_column(self, column_id: str, table_id: str, changes: dict) -> ColumnDTO:
        raise NotImplementedError()


class ProductionTablesApiHttpClient(TablesApiHttpClient):
    def execute(self, query: str, params: typing.List) -> requests.Response:
        body = {"query": query, "params": params}
        return requests.post(
            self.execute_url,
            headers=SIDECAR_HEADERS,
            json=body,
            timeout=REQUEST_TIMEOUT,
        )

    def create_table(self, id: str, name: str) -> TableDTO:
        raise NotImplementedError()

    def create_column(self, table_id: str, name: str, type: str) -> ColumnDTO:
        raise NotImplementedError()

    def insert_row(self, table_id: str, values: dict):
        raise NotImplementedError()

    def update_table(self, table_id: str, name: str) -> TableDTO:
        raise NotImplementedError()

    def update_column(self, column_id: str, table_id: str, changes: dict) -> ColumnDTO:
        raise NotImplementedError()


class LocalTablesApiHttpClient(TablesApiHttpClient):
    def _request(
        self,
        method: str,
        path: str,
        body: Any = None,
        params: dict = {},
        raise_for_status: bool = True,
    ):
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to execute a table query")
        r = requests.request(
            method=method,
            url=f"{self.base_url}/tables{path}",
            headers=headers,
            json=body,
            params=params,
            timeout=REQUEST_TIMEOUT,
        )

        if raise_for_status:
            r.raise_for_status()

        return r

    def execute(self, query: str, params: typing.List) -> requests.Response:
        body = {"query": query, "params": params}
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to execute a table query")
        return requests.post(
            self.execute_url, headers=headers, json=body, timeout=REQUEST_TIMEOUT
        )

    def create_table(self, id: str, name: str) -> TableDTO:
        r = self._request(
            "POST",
            path="/table",
            body={
                "id": id,
                "name": name,
            },
        )
        table = r.json()["response"]["table"]
        return TableDTO.from_dict(table)

    def create_column(self, table_id: str, name: str, type: str) -> ColumnDTO:
        body = {
            "tableId": table_id,
            "name": name,
            "type": type,
        }
        r = self._request("POST", "/column", body=body)
        column = r.json()["response"]
        return ColumnDTO.from_dict(column)

    def insert_row(self, table_id: str, values: dict):
        body = {
            "tableId": table_id,
            "row": values,
        }
        self._request("POST", "/row", body=body)

    def update_table(self, table_id: str, name: str) -> TableDTO:
        body = {
            "name": name,
        }
        r = self._request("PATCH", f"/table/{table_id}", body=body)
        table = r.json()["response"]
        return TableDTO.from_dict(table)

    def update_column(self, column_id: str, table_id: str, changes: dict) -> ColumnDTO:
        body = {
            "tableId": table_id,
            "changes": changes,
        }
        r = self._request("PATCH", f"column/{column_id}", body=body)
        column = r.json()["response"]
        return ColumnDTO.from_dict(column)



import importlib
import sys
import threading
import time
from pathlib import Path
from typing import Optional

from dotenv import load_dotenv
from watchdog.events import FileModifiedEvent, FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.settings import Settings

IGNORED_FILES = [".abstra/resources.dat", ".abstra/", "abstra.json", "__pycache__"]
DEBOUNCE_DELAY = 0.5


class FileChangeEventHandler(FileSystemEventHandler):
    def __init__(self, project_repository: ProjectRepository):
        super().__init__()
        self._debounce_timer: Optional[threading.Timer] = None
        self.dot_env_path = Settings.root_path / ".env"
        self.project_repository = project_repository

    def on_modified(self, event: FileSystemEvent):
        filepath = Path(event.src_path)

        if self.should_ignore_event(event) or self.should_ignore_path(filepath):
            return

        if filepath.resolve() == self.dot_env_path.resolve():
            AbstraLogger.info("Reloading .env and all modules")
            load_dotenv(self.dot_env_path, override=True)
            for dep in self.project_repository.load().get_local_dependencies():
                self.reload_module(dep)
            return

        resolved_deps = [
            dep.resolve()
            for dep in self.project_repository.load().get_local_dependencies()
        ]

        if filepath.resolve() in resolved_deps:
            AbstraLogger.info(f"Reloading modified module: {filepath}")
            self.reload_module(filepath)
            return

    def reload_module(self, file: Path):
        module_name = file.stem
        module = sys.modules.get(module_name)

        try:
            if module is None:
                importlib.import_module(module_name)
            else:
                if module.__spec__ is not None and module.__spec__.cached is not None:
                    Path(module.__spec__.cached).unlink(missing_ok=True)
                importlib.reload(module)
        except Exception as e:
            AbstraLogger.error(
                f"Could not reload module from {file} with the following error: {e}"
            )

    def should_ignore_path(self, path: Path) -> bool:
        return any(ignored_file in str(path) for ignored_file in IGNORED_FILES)

    def should_ignore_event(self, event: FileSystemEvent) -> bool:
        return not isinstance(event, (FileModifiedEvent))

    def run(self):
        observer = Observer()
        observer.schedule(self, path=str(Settings.root_path), recursive=True)
        observer.start()

        try:
            while True:
                time.sleep(1)

        except KeyboardInterrupt:
            observer.stop()
        observer.join()



from typing import Any, List, Optional, Union

from abstra_internals.entities.forms.widgets.widget_base import InputWidget


class TagInput(InputWidget):
    """Tag input widget for entering multiple tag values.

    Attributes:
        value (List[Any]): The list of tag values entered by the user.
    """

    type = "tag-input"
    value: List[Any]

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        placeholder: str = "",
        required: bool = True,
        hint: Optional[str] = None,
        full_width: bool = False,
        disabled: bool = False,
        errors: Optional[Union[List[str], str]] = None,
    ):
        """Initialize a TagInput widget.

        Args:
            label (str): Text label displayed above the input.
            key (Optional[str]): Identifier for the widget, defaults to label if not provided.
            placeholder (str): Placeholder text displayed when the input is empty.
            required (bool): Whether at least one tag must be entered before proceeding.
            hint (Optional[str]): Help text displayed below the input.
            full_width (bool): Whether the input should take up the full width of its container.
            disabled (bool): Whether the input is non-interactive.
            errors (Optional[Union[List[str], str]]): Pre-defined validation error messages to display.
        """
        self.label = label
        self.key = key or label
        self.placeholder = placeholder
        self.required = required
        self.hint = hint
        self.full_width = full_width
        self.disabled = disabled
        self.value = []
        self.errors = errors

    def is_value_unset(self):
        return False

    def _render(self):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "value": self.value,
            "placeholder": self.placeholder,
            "required": self.required,
            "hint": self.hint,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "errors": self.errors,
        }



from unittest import TestCase

from abstra_internals.repositories.project.json_migrations.migration_007 import (
    Migration007,
)


class TestMigration07(TestCase):
    def test_adds_item_name_to_iterators(self):
        m = Migration007(
            {
                "version": "6.0",
                "iterators": [
                    {
                        "id": "b35faa65-5910-4cd4-b56c-c2bee99b9263",
                        "title": "iter",
                        "variable_name": "mylist",
                        "workflow_position": [-256, -32],
                        "transitions": [],
                    }
                ],
            }
        )
        m.apply()
        self.assertEqual(
            m.data,
            {
                "version": "7.0",
                "iterators": [
                    {
                        "id": "b35faa65-5910-4cd4-b56c-c2bee99b9263",
                        "title": "iter",
                        "variable_name": "mylist",
                        "workflow_position": [-256, -32],
                        "transitions": [],
                        "item_name": "item",
                    }
                ],
            },
        )

    def test_adds_only_to_iterators(self):
        m = Migration007(
            {
                "version": "6.0",
                "conditions": [{"title": "it must be mantained"}],
                "iterators": [
                    {
                        "id": "b35faa65-5910-4cd4-b56c-c2bee99b9263",
                        "title": "iter",
                        "variable_name": "mylist",
                        "workflow_position": [-256, -32],
                        "transitions": [],
                        "item_name": "item",
                    }
                ],
            }
        )
        m.apply()
        self.assertEqual(
            m.data,
            {
                "version": "7.0",
                "conditions": [{"title": "it must be mantained"}],
                "iterators": [
                    {
                        "id": "b35faa65-5910-4cd4-b56c-c2bee99b9263",
                        "title": "iter",
                        "variable_name": "mylist",
                        "workflow_position": [-256, -32],
                        "transitions": [],
                        "item_name": "item",
                    }
                ],
            },
        )

    def test_it_should_not_break_with_no_iterators(self):
        m = Migration007({"version": "6.0"})
        m.apply()
        self.assertEqual(m.data, {"version": "7.0"})



import mimetypes
from datetime import datetime
from os.path import sep
from pathlib import Path
from typing import List, Optional

import flask

from abstra_internals.contracts_generated import (
    AbstraLibApiEditorFilesDeleteResponse,
    AbstraLibApiEditorFilesEditResponse,
    AbstraLibApiEditorFilesListResponse,
    AbstraLibApiEditorFilesListResponseItem,
    AbstraLibApiEditorFilesListResponseItemStagesItem,
    AbstraLibApiEditorFilesMkdirResponse,
    AbstraLibApiEditorFilesRenameResponse,
    AbstraLibApiEditorFilesSettingsResponse,
    CommonFileNode,
)
from abstra_internals.repositories.factory import Repositories


def rm_tree(pth: Path):
    pth = Path(pth)
    if not pth.exists():
        return
    for child in pth.glob("*"):
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)
    pth.rmdir()


class CodebaseController:
    def __init__(self, repos: Repositories):
        self.repos = repos

    def list_files(self, path) -> AbstraLibApiEditorFilesListResponse:
        if path is None:
            path = Path.cwd()
        elif isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")

        project = self.repos.project.load()

        return [
            AbstraLibApiEditorFilesListResponseItem(
                file=CommonFileNode(
                    path_parts=list(child_path.relative_to(path).parts),
                    size=child_path.stat().st_size,
                    last_modified=datetime.fromtimestamp(path.stat().st_mtime),
                    type="directory" if child_path.is_dir() else "file",
                    children=(
                        [list(c.relative_to(path).parts) for c in child_path.iterdir()]
                        if child_path.is_dir()
                        else []
                    ),
                ),
                stages=[
                    AbstraLibApiEditorFilesListResponseItemStagesItem(
                        id=stage.id,
                        type=stage.type_name,  # type: ignore
                    )
                    for stage in project.get_stages_by_file_path(child_path)
                ],
            )
            for child_path in path.glob("**/*")
        ]

    def create_file(self, path, content: Optional[bytes] = None) -> CommonFileNode:
        if isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")

        if content is not None:
            path.write_bytes(content)
        else:
            path.touch()
        return CommonFileNode(
            path_parts=list(path.parts),
            size=path.stat().st_size,
            last_modified=datetime.fromtimestamp(path.stat().st_mtime),
            type="directory" if path.is_dir() else "file",
        )

    def delete_file(
        self, path_parts: List[str]
    ) -> AbstraLibApiEditorFilesDeleteResponse:
        path = Path(*path_parts)
        if path.is_dir():
            rm_tree(path)
        else:
            path.unlink()
        return AbstraLibApiEditorFilesDeleteResponse(ok=True)

    def rename_file(self, path, new_name) -> AbstraLibApiEditorFilesRenameResponse:
        if isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")

        new_path = path.parent / new_name
        path.rename(new_path)
        return AbstraLibApiEditorFilesRenameResponse(ok=True)

    def edit_file(self, path, content: bytes) -> AbstraLibApiEditorFilesEditResponse:
        if isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")

        path.write_bytes(content)
        return AbstraLibApiEditorFilesEditResponse(ok=True)

    def get_file(self, path):
        if isinstance(path, str):
            path = Path(path).absolute()
        elif isinstance(path, Path):
            path = path.absolute()
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")
        mtype, _ = mimetypes.guess_type(path)
        return flask.send_file(path, mimetype=mtype)

    def mkdir(self, path) -> AbstraLibApiEditorFilesMkdirResponse:
        if isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")

        path.mkdir(parents=True, exist_ok=True)
        return AbstraLibApiEditorFilesMkdirResponse(ok=True)

    def settings(self) -> AbstraLibApiEditorFilesSettingsResponse:
        return AbstraLibApiEditorFilesSettingsResponse(
            separator=sep,
        )



import flask

from abstra_internals.controllers.agents import AgentsController
from abstra_internals.controllers.main import MainController


def get_editor_bp(main_controller: MainController):
    bp = flask.Blueprint("editor_agent", __name__)
    controller = AgentsController(main_controller)

    @bp.get("/connections")
    def _get_agents():
        return controller.get_agent_connections()

    @bp.get("/is-usage-mode")
    def _get_is_usage_mode():
        return controller.get_is_usage_mode()

    return bp



from .agent import get_player_bp as role_agent_player_bp
from .client import get_editor_bp as role_client_editor_bp
from .client import get_player_bp as role_client_player_bp

__all__ = ["role_agent_player_bp", "role_client_player_bp", "role_client_editor_bp"]



from abstra_internals.entities.forms.widgets.widget_base import OutputWidget


class HtmlOutput(OutputWidget):
    """HTML output widget for displaying custom HTML content.

    Attributes:
        html (str): The HTML content to display.
    """

    type = "html-output"

    def __init__(
        self,
        html: str,
        *,
        full_width: bool = False,
    ):
        """Initialize an HtmlOutput widget.

        Args:
            html (str): The HTML content to display.
            full_width (bool): Whether the content should take up the full width of its container.
        """
        self.html = html
        self.full_width = full_width

    def _render(self):
        return {"type": self.type, "html": self.html, "fullWidth": self.full_width}



from .FileInput import FileInput


class VideoInput(FileInput):
    """Video upload input widget for handling video file uploads.

    Attributes:
        value (Union[FileResponse, List[FileResponse], None]): The uploaded video file(s).
    """

    type = "video-input"



import unittest
from io import BytesIO
from pathlib import Path
from unittest.mock import MagicMock, patch

from PIL import Image

from abstra_internals.controllers.sdk.sdk_ai import AiSDKController
from abstra_internals.interface.sdk.forms.deprecated.widgets.response_types import (
    FileResponse,
)


class TestAiSDKController(unittest.TestCase):
    def setUp(self):
        # Mock the AiApiHttpClient to isolate tests from external API calls
        self.mock_ai_client = MagicMock()
        self.controller = AiSDKController(self.mock_ai_client)

    def test_extract_pdf_images(self):
        # Mock pdfium.PdfDocument to return a mock page
        mock_page = MagicMock()
        mock_page.render.return_value.to_pil.return_value = Image.new("RGB", (100, 100))
        mock_pdf = MagicMock()
        mock_pdf.__iter__.return_value = [mock_page]

        with patch("pypdfium2.PdfDocument", return_value=mock_pdf):
            images = self.controller._extract_pdf_images(mock_pdf)

        self.assertEqual(len(images), 1)  # One image is expected
        self.assertTrue(
            isinstance(images[0], BytesIO)
        )  # Image should be in BytesIO format

    def test_make_image_url_message(self):
        url = "http://example.com/image.png"
        expected_message = {
            "role": "user",
            "content": [
                {
                    "type": "image_url",
                    "image_url": {
                        "url": url,
                    },
                },
            ],
        }

        result = self.controller._make_image_url_message(url)
        self.assertEqual(result, expected_message)

    def test_make_text_message(self):
        text = "Some text"
        expected_message = {
            "role": "user",
            "content": text,
        }

        result = self.controller._make_text_message(text)
        self.assertEqual(result, expected_message)

    def test_make_messages_from_str(self):
        text = "This is a prompt"
        result = self.controller._make_messages(text)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["role"], "user")
        self.assertEqual(result[0]["content"], text)

    def test_make_messages_from_image(self):
        # Create a dummy image to simulate the prompt
        img = Image.new("RGB", (100, 100))
        image_io = BytesIO()
        img.save(image_io, format="PNG")
        image_io.seek(0)

        result = self.controller._make_messages(image_io)
        self.assertEqual(len(result), 1)
        self.assertIn("type", result[0]["content"][0])
        self.assertEqual(result[0]["content"][0]["type"], "image_url")

    def test_make_messages_from_path(self):
        file_path = Path("test.pdf")
        file_path.write_text("dummy content")  # Creating a dummy file

        with patch("builtins.open", MagicMock(return_value=BytesIO(b"fake content"))):
            result = self.controller._make_messages(file_path)

        self.assertEqual(len(result), 1)
        self.assertIn("type", result[0]["content"][0])

    def test_prompt_with_image_file(self):
        # Mock AI client response
        self.mock_ai_client.prompt.return_value = {"content": "Mocked AI response"}

        # Create a dummy image
        img = Image.new("RGB", (100, 100))
        image_io = BytesIO()
        img.save(image_io, format="PNG")
        image_io.seek(0)

        result = self.controller.prompt(
            prompts=[image_io],
            instructions=["This is a test instruction."],
            format=None,
            temperature=0.7,
        )

        self.assertEqual(result, "Mocked AI response")

    def test_prompt_with_file_response(self):
        # Mock AI client response
        self.mock_ai_client.prompt.return_value = {"content": "Mocked AI response"}

        # save dummy file
        file_path = Path("test.pdf")
        file_path.write_text("dummy content")

        class MockFileResponse(FileResponse):
            def __init__(self, url):
                self._url = url

            @property
            def name(self):
                return self.path.name

            @property
            def content(self):
                return self.path.read_bytes()

            @property
            def file(self):
                return self.path.open("rb")

            @property
            def path(self):
                return file_path

        # Mock FileResponse
        file_response = MockFileResponse(url=file_path.as_posix())

        result = self.controller.prompt(
            prompts=[file_response],
            instructions=["Test instruction."],
            format=None,
            temperature=0.7,
        )

        self.assertEqual(result, "Mocked AI response")



import inspect
import signal

from gunicorn.arbiter import Arbiter
from gunicorn.workers.base import Worker

from abstra_internals.controllers.main import MainController
from abstra_internals.environment import (
    DEFAULT_PORT,
    THREADS,
    WORKER_CLASS,
    WORKER_TEMP_DIR,
    WORKERS,
    set_SERVER_UUID,
    set_WORKER_UUID,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.utils import get_internal_id


class GunicornOptionsBuilder:
    def __init__(self, main_controller: MainController) -> None:
        self.main_controller = main_controller

    #### HOOKS

    # Called just before the master process is initialized.
    def on_starting(self, server: Arbiter):
        server_uuid = get_internal_id(server)
        set_SERVER_UUID(server_uuid)

    # Called just before a worker is forked.
    def pre_fork(self, server: Arbiter, worker: Worker):
        worker_uuid = get_internal_id(worker)
        set_WORKER_UUID(worker_uuid)

    # Called just after a worker has been exited, in the master process.
    def child_exit(self, server: Arbiter, worker: Worker):
        try:
            # gambi do gunicorn
            # https://github.com/benoitc/gunicorn/blob/master/gunicorn/arbiter.py#L515
            status = -1
            if curr_frame := inspect.currentframe():
                if back := curr_frame.f_back:
                    back_status = back.f_locals.get("status")
                    if isinstance(back_status, int):
                        status = back_status

            worker_id = get_internal_id(worker, ensure=False)
            app_id = get_internal_id(server, ensure=False)

            err_msg = f"[ABORTED] Worker exited with status ({status})"
            if status == signal.Signals.SIGKILL:
                err_msg += ": Server reached its memory limit or was replaced with a new version"

            self.main_controller.worker_exit(
                app_id=app_id, worker_id=worker_id, err_msg=err_msg
            )

        except Exception as e:
            AbstraLogger.capture_exception(e)

    # Called just before exiting Gunicorn.
    def on_exit(self, server: Arbiter):
        try:
            app_id = get_internal_id(server, ensure=False)
            self.main_controller.app_exit(
                app_id=app_id, err_msg="[ABORTED] Server exited"
            )

        except Exception as e:
            AbstraLogger.capture_exception(e)

    def build(self):
        return {
            "bind": f":{DEFAULT_PORT or 8002}",
            "workers": WORKERS,
            "threads": THREADS,
            "worker_class": WORKER_CLASS,
            "worker_tmp_dir": WORKER_TEMP_DIR,
            "on_starting": self.on_starting,
            "pre_fork": self.pre_fork,
            "child_exit": self.child_exit,
            "on_exit": self.on_exit,
        }



import unittest
from typing import Generator, List, Optional
from unittest.mock import patch

from abstra_internals.controllers.sdk.sdk_context import SDKContextStore
from abstra_internals.controllers.sdk.sdk_forms import FormSDKController
from abstra_internals.entities.forms.form_state import State
from abstra_internals.entities.forms.steps import (
    ComputationStep,
    GeneratorStep,
    PageStep,
    Step,
)
from abstra_internals.entities.forms.template import Template
from abstra_internals.entities.forms.widgets.library import TextOutput
from abstra_internals.entities.forms.widgets.widget_base import Widget

from .form import Runnable, run


def page_function(state: State) -> Template:
    return [TextOutput("Hello, world!")]


def computation_function(state: State):
    state["a"] = 1 + 1


def page_function_with_optional_return(state: State) -> Optional[Template]:
    if state.get("a"):
        return [TextOutput("Hello, world!")]


def generator_function(state: State) -> Generator[Template, None, None]:
    yield [TextOutput("Hello, world!")]


basic_template: List[Widget] = [TextOutput("Hello, world!")]


class MockFormSDKController(FormSDKController):
    def __init__(self) -> None:
        pass

    def run_form(self, *, steps: List[Step], state: State, hide_steps: bool) -> State:
        return State({"mock": "mock"})


class MockContext:
    def __init__(self) -> None:
        self.form_sdk = MockFormSDKController()


class FormInterfaceTest(unittest.TestCase):
    @patch.object(SDKContextStore, "get_by_thread", return_value=MockContext())
    def test_multiple_inputs(self, mock_get_by_thread):
        with patch.object(MockFormSDKController, "run_form") as mock_run_form:
            # mock SDKContextStore.get_by_thread().form_sdk
            runnables: List[Runnable] = [
                page_function,
                computation_function,
                page_function_with_optional_return,
                generator_function,
                basic_template,
            ]

            run(runnables)

            classes = [type(arg) for arg in mock_run_form.call_args_list[0][1]["steps"]]

            self.assertEqual(
                classes, [PageStep, ComputationStep, PageStep, GeneratorStep, PageStep]
            )

    @patch.object(SDKContextStore, "get_by_thread", return_value=MockContext())
    def test_single_widget(self, mock_get_by_thread):
        with patch.object(MockFormSDKController, "run_form") as mock_run_form:
            # mock SDKContextStore.get_by_thread().form_sdk

            run(TextOutput("Hello, world!"))

            classes = [type(arg) for arg in mock_run_form.call_args_list[0][1]["steps"]]

            self.assertEqual(classes, [PageStep])



from typing import Optional

import requests

from abstra_internals.email_templates import task_waiting_template
from abstra_internals.entities.execution import Execution
from abstra_internals.entities.execution_context import ScriptContext
from abstra_internals.environment import IS_PRODUCTION, REQUEST_TIMEOUT
from abstra_internals.repositories.factory import Repositories
from abstra_internals.repositories.project.project import (
    AgentStage,
    ClientStage,
    FormStage,
    ScriptStage,
    Stage,
)
from abstra_internals.repositories.tasks import TaskDTO, TaskPayload


class TaskExecutor:
    def __init__(self, repos: Repositories) -> None:
        self.repos = repos
        self.project = self.repos.project.load()

    def send_task(
        self,
        type: str,
        current_stage: Stage,
        payload: TaskPayload,
        execution: Optional[Execution] = None,
        show_warning: bool = True,
    ) -> None:
        project = self.repos.project.load()
        next_stages = [
            project.get_stage_raises(t.target_id)
            for t in current_stage.workflow_transitions
            if t.matches(type)
        ]

        if len(next_stages) == 0 and show_warning:
            print(
                f"[WARNING] No transitions found for task type {type} in stage {current_stage.id}"
            )
            return

        for stage in next_stages:
            task = self.repos.tasks.send_task(
                type=type,
                payload=payload,
                source_stage_id=current_stage.id,
                target_stage_id=stage.id,
                execution_id=execution.id if execution else None,
            )
            self._send_waiting_thread_notification(task)
            if execution:
                execution.context.sent_tasks.append(task.id)
            if isinstance(stage, ScriptStage):
                self.repos.producer.enqueue(
                    context=ScriptContext(task_id=task.id),
                    stage_id=stage.id,
                )
            elif (
                isinstance(stage, AgentStage)
                and stage.project_id is not None
                and stage.client_stage_id is not None
            ):
                agent = self.repos.role_clients.get_agent(stage.project_id)

                conn = next(
                    (
                        c
                        for c in self.repos.role_clients.get_connections()
                        if c.agent_project_id == agent.project_id
                        and c.client_stage_id == stage.id
                    ),
                    None,
                )

                assert conn is not None, "Connection for agent not found"

                requests.post(
                    agent.tasks_url + "/agent",
                    json={
                        "task_data": {
                            "type": type,
                            "payload": {
                                **payload,
                                "connection_token": conn.token,
                            },
                        },
                        "target_stage_id": (stage.client_stage_id),
                        "execution_id": (
                            execution.id if IS_PRODUCTION and execution else None
                        ),
                    },
                    headers={"authorization": conn.token},
                    timeout=REQUEST_TIMEOUT,
                ).raise_for_status()

            elif isinstance(stage, ClientStage):
                assert isinstance(payload["connection_token"], str)
                conn = self.repos.role_agents.get_connection_by_token(
                    payload["connection_token"]
                )

                requests.post(
                    conn.client_task_url,
                    json=task.model_dump(),
                    headers={"Authorization": conn.token},
                    timeout=REQUEST_TIMEOUT,
                ).raise_for_status()

    def _send_waiting_thread_notification(self, task: TaskDTO):
        stage = self.project.get_stage(task.target_stage_id)
        if not stage:
            raise Exception(f"Stage {task.target_stage_id} not found")

        if not (isinstance(stage, FormStage) and stage.notification_trigger.enabled):
            return

        recipient_emails = stage.notification_trigger.get_recipients(task.payload)
        if not recipient_emails:
            return

        self.repos.email.send(
            task_waiting_template.generate_email(
                recipient_emails=recipient_emails, form=stage
            )
        )



import io
from typing import List, Union

from abstra_internals.controllers.sdk.sdk_context import SDKContextStore
from abstra_internals.email_templates import message_template


def send_email(
    to: Union[str, List[str]],
    message: str,
    title: str = "",
    attachments: List[Union[str, io.IOBase]] = [],
    is_html: bool = False,
):
    """Send an email to one or more recipients.

    Args:
        to (Union[str, List[str]]): Email address(es) of the recipient(s). Can be a single
            email address as a string or multiple addresses as a list of strings.
        message (str): Content of the email message.
        title (str): Subject line of the email. Defaults to "".
        attachments (List[Union[str, io.IOBase]]): List of attachments to include.
            Each attachment can be a file path string or an open file-like object. Defaults to [].
        is_html (bool): Whether the message content is HTML. When True, the message
            will be sent as HTML; when False, the message will be sent as plain text. Defaults to False.
    """
    if isinstance(to, str):
        to = [to]
    SDKContextStore.get_by_thread().repositories.email.send(
        message_template.generate_email(to, message, title, attachments, is_html)
    )


__all__ = [
    "send_email",
]



import copy
import json
from pathlib import Path
from unittest import TestCase

from abstra_internals.repositories.project.json_migrations.migration_012 import (
    Migration012,
)


class TestMigration12(TestCase):
    def test_base_json(self):
        base_abstra_json = Path(__file__).parent / "json_tests" / "base.json"
        json_data = json.loads(base_abstra_json.read_text())
        expected_result = copy.deepcopy(json_data)
        m = Migration012(json_data)
        m.set_as_test()
        m.apply()
        del expected_result["conditions"]
        del expected_result["iterators"]
        del expected_result["kanban"]
        expected_result["version"] = "12.0"
        self.assertEqual(
            m.data,
            expected_result,
        )

    def test_simple_json(self):
        simple_abstra_json = Path(__file__).parent / "json_tests" / "simple.json"
        json_data = json.loads(simple_abstra_json.read_text())
        expected_result = copy.deepcopy(json_data)
        m = Migration012(json_data)
        m.set_as_test()
        m.apply()
        del expected_result["conditions"]
        del expected_result["iterators"]
        del expected_result["kanban"]
        for transition in expected_result["forms"][0]["transitions"]:
            transition["type"] = "task"
            transition["task_type"] = "success"
            del transition["condition_value"]
        expected_result["version"] = "12.0"
        self.assertEqual(
            m.data,
            expected_result,
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

import pika

from abstra_internals.entities.execution_context import ClientContext
from abstra_internals.environment import (
    RABBITMQ_DEFAUT_EXCHANGE,
    RABBITMQ_EXECUTION_QUEUE,
)
from abstra_internals.repositories.multiprocessing import MPContext
from abstra_internals.utils.serializable import Serializable


class PreExecution(Serializable):
    stage_id: str
    context: Optional[ClientContext] = None


@dataclass
class QueueMessage:
    preexecution: PreExecution
    delivery_tag: int


class ProducerRepository(ABC):
    @abstractmethod
    def enqueue(self, stage_id: str, context: Optional[ClientContext] = None) -> None:
        raise NotImplementedError()


class LocalProducerRepository(ProducerRepository):
    def __init__(self, mp_context: MPContext):
        self.queue = mp_context.Queue()

    def enqueue(self, stage_id: str, context: Optional[ClientContext] = None) -> None:
        preexecution = PreExecution(stage_id=stage_id, context=context)
        self.queue.put(
            QueueMessage(
                preexecution=preexecution,
                delivery_tag=0,
            ),
        )


class ProductionProducerRepository(ProducerRepository):
    def __init__(self, connection_uri: str):
        self.connection_uri = connection_uri

    @property
    def conn_params(self):
        return pika.URLParameters(self.connection_uri)

    @property
    def props(self):
        return pika.BasicProperties(
            delivery_mode=pika.DeliveryMode.Persistent,
            content_type="application/json",
        )

    def enqueue(self, stage_id: str, context: Optional[ClientContext] = None) -> None:
        preexecution = PreExecution(stage_id=stage_id, context=context)
        with pika.BlockingConnection(self.conn_params) as connection:
            with connection.channel() as channel:
                channel.queue_declare(queue=RABBITMQ_EXECUTION_QUEUE, durable=True)
                channel.basic_publish(
                    body=preexecution.model_dump_json(),
                    routing_key=RABBITMQ_EXECUTION_QUEUE,
                    exchange=RABBITMQ_DEFAUT_EXCHANGE,
                    properties=self.props,
                )



import ast
import webbrowser
from typing import Dict, List

from abstra_internals.constants import (
    get_project_url,
)
from abstra_internals.linter.linter import LinterFix, LinterIssue, LinterRule
from abstra_internals.repositories.project.project import LocalProjectRepository

DEPRECATED_FUNCTIONS = {
    "run": {"from": "abstra.tables", "new_function": "run_sql"},
}


class ReplaceDeprecatedFunction(LinterFix):
    def __init__(self, stage_type: str, stage_id: str):
        self.label = "Go to the file"
        self.stage_type = stage_type
        self.stage_id = stage_id

    def fix(self):
        project_url = get_project_url()
        webbrowser.open(f"{project_url}/_editor/{self.stage_type}/{self.stage_id}")


class DeprecatedFunctionFound(LinterIssue):
    def __init__(
        self,
        function_name: str,
        new_function_name: str,
        stage_type: str,
        stage_id: str,
        module: str,
    ):
        self.label = f"The function '{function_name}' from '{module}' is deprecated. Please use '{new_function_name}' instead."
        self.fixes = [ReplaceDeprecatedFunction(stage_type, stage_id)]


class DeprecatedFunctionUsage(LinterRule):
    label: str = "Deprecated function usage"
    type: str = "warning"

    def find_issues(self) -> List[LinterIssue]:
        project = LocalProjectRepository().load()
        issues = []

        for entrypoint, stage in project.iter_entrypointed_stages():
            try:
                with open(entrypoint, "r", encoding="utf-8") as file:
                    source_code = file.read()
                    tree = ast.parse(source_code)

                    import_map = self.track_imports(tree)

                    for node in ast.walk(tree):
                        if isinstance(node, ast.Call):
                            if isinstance(node.func, ast.Name):
                                function_name = node.func.id
                                if (
                                    function_name in DEPRECATED_FUNCTIONS
                                    and DEPRECATED_FUNCTIONS[function_name]["from"]
                                    == import_map.get(function_name)
                                ):
                                    new_function_name = DEPRECATED_FUNCTIONS[
                                        function_name
                                    ]["new_function"]
                                    module = DEPRECATED_FUNCTIONS[function_name]["from"]
                                    issues.append(
                                        DeprecatedFunctionFound(
                                            function_name,
                                            new_function_name,
                                            stage_type=stage.type_name,
                                            stage_id=stage.id,
                                            module=module,
                                        )
                                    )

            except Exception as e:
                print(f"Error while processing {entrypoint}: {e}")

        return issues

    def track_imports(self, tree: ast.AST) -> Dict[str, str]:
        import_map = {}

        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for alias in node.names:
                    import_map[alias.asname or alias.name.split(".")[-1]] = alias.name

            elif isinstance(node, ast.ImportFrom):
                module_name = node.module
                for alias in node.names:
                    import_map[alias.asname or alias.name] = module_name

        return import_map



from abc import ABC, abstractmethod
from typing import Callable, Generic, List, Optional, TypedDict, TypeVar, Union, final


class Widget(ABC):
    type: str

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def _render(self):
        raise NotImplementedError("render method not implemented")


class OutputWidget(Widget):
    pass


WidgetValue = TypeVar("WidgetValue")


class InputWidget(Widget, Generic[WidgetValue]):
    value: WidgetValue
    required: bool
    key: Optional[str]
    _errors: Optional[List[str]]

    @final
    @property
    def errors(self) -> List[str]:
        return self._errors or []

    @final
    @errors.setter
    def errors(self, value: Union[List[str], str, None]) -> None:
        if value is None:
            self._errors = []
            return

        if isinstance(value, str):
            self._errors = [value]
            return

        if isinstance(value, list):
            if all(isinstance(e, str) for e in value):
                self._errors = value
                return
            else:
                raise TypeError("Provided errors list contains non-string elements")

        raise TypeError("Errors must be a list of strings or a string")

    @final
    def _safe_get_key(self, idx: int) -> str:
        return self.key or f"{self.type}_{idx}"

    @final
    def _apply_validation(self):
        for validator in self._validators:
            self.errors += validator()

    @property
    def _validators(self) -> List[Callable[[], List[str]]]:
        return [self._basic_validate_required]

    @final
    def _basic_validate_required(self) -> List[str]:
        if self.required and (self.value is None or self.value == ""):
            return ["i18n_error_required_field"]

        return []

    def _parse_value(self, value):
        return value


class LabelValueOption(TypedDict):
    label: str
    value: object


AbstraOption = Union[str, LabelValueOption]


class OptionsHandler:
    def __init__(self, options: List["AbstraOption"]) -> None:
        self.options = options

        self._mappedOptions = {
            str(index): {
                "label": option if isinstance(option, str) else option["label"],
                "value": option if isinstance(option, str) else option["value"],
            }
            for index, option in enumerate(options)
        }

    def ids_from_values(self, values: List[object]) -> List[str]:
        value = [self._id_from_value(value) for value in values]
        return [v for v in value if v is not None]

    def values_from_ids(self, ids: Union[List[str], None]) -> List[object]:
        if ids is None:
            return []

        return [self._value_from_id(id) for id in ids]

    def _value_from_id(self, id: str) -> object:
        if id not in self._mappedOptions:
            return None
        return self._mappedOptions[id]["value"]

    def _id_from_value(self, value: object) -> Optional[str]:
        for mapKey, mapValue in self._mappedOptions.items():
            if mapValue["value"] == value:
                return mapKey
        return None

    def serialized_options(self) -> List[LabelValueOption]:
        return [
            {"label": v["label"], "value": k} for k, v in self._mappedOptions.items()
        ]


class MultipleHandler:
    def __init__(
        self,
        multiple: bool,
        min: Optional[int] = None,
        max: Optional[int] = None,
        required: Optional[Union[bool, str]] = None,
    ) -> None:
        self.multiple = multiple
        self.min = min
        self.max = max
        self.required = required

    def validate(self, value: Optional[Union[object, List[object]]]) -> List[str]:
        if (value is None or value == []) and self.required:
            return ["i18n_error_required_field"]

        if not self.multiple or not isinstance(value, list):
            return []

        if self.min is not None and len(value) < self.min:
            return ["i18n_error_min_list"]

        if self.max is not None and len(value) > self.max:
            return ["i18n_error_max_list"]

        return []

    T = TypeVar("T", bound=object)

    def value_to_list_or_value(self, value: List[T]) -> Union[List[T], T, None]:
        if not isinstance(value, list) and value is not None:
            raise Exception("Non-list value received")

        if self.multiple:
            return value

        return value[0] if len(value) else None

    def value_to_list(self, value: Union[List[T], T, None]) -> List[T]:
        if isinstance(value, list):
            return value
        if value is None:
            return []
        return [value]


class NumberValueHandler:
    def __init__(
        self,
        min: Optional[Union[int, float]] = None,
        max: Optional[Union[int, float]] = None,
        required: Optional[Union[bool, str]] = None,
    ) -> None:
        self.min = min
        self.max = max
        self.required = required

    def validate(self, value: Optional[Union[int, float]]) -> List[str]:
        errors = []
        if self.required and value is None:
            if isinstance(self.required, str):
                errors.append(self.required)
            else:
                errors.append("i18n_error_required_field")

        if self.min is not None and value is not None and value < self.min:
            errors.append("i18n_error_min_number")

        if self.max is not None and value is not None and value > self.max:
            errors.append("i18n_error_max_number")

        return errors



from os.path import sep
from pathlib import Path


class NotSubpathError(Exception):
    def __init__(self) -> None:
        super().__init__("Path is not a subpath of the base path")


class NotAbsoluteBasePathError(Exception):
    def __init__(self) -> None:
        super().__init__("Base path must be absolute if path is absolute")


def get_relative_path(path: Path, base_path: Path) -> Path:
    if path.is_absolute() and base_path.is_absolute():
        if str(path).startswith(str(base_path)):
            relative_part = str(path).replace(str(base_path) + sep, "")
            return Path(relative_part)
        else:
            raise NotSubpathError()
    elif path.is_absolute() and not base_path.is_absolute():
        raise NotAbsoluteBasePathError()
    elif not path.is_absolute() and base_path.is_absolute():
        return path
    else:
        path_str = str(path)
        base_path_str = str(base_path)
        if path_str.startswith(base_path_str):
            relative_part = path_str.replace(base_path_str + sep, "")
            return Path(relative_part)
        else:
            raise NotSubpathError()



from typing import List, Literal

from abstra_internals.controllers.sdk.sdk_context import SDKContextStore
from abstra_internals.entities.agents import ConnectionModel


def get_connections(role: Literal["client", "agent"]) -> List[ConnectionModel]:
    """
    Get a list of available connections based on the specified role.

    Args:
        role (Literal["client", "agent"]): The role to get connections for, either "client" or "agent".

    Returns:
        List[ConnectionModel]: A list of connection models for the specified role.
    """
    if role == "client":
        return (
            SDKContextStore.get_by_thread().repositories.role_clients.get_connections()
        )
    else:
        return (
            SDKContextStore.get_by_thread().repositories.role_agents.get_connections()
        )


def validate_token(token: str):
    return SDKContextStore.get_by_thread().repositories.agents.validate_token(token)


def accept_connection(token: str):
    return SDKContextStore.get_by_thread().repositories.agents.accept_connection(token)


__all__ = ["get_connections"]



import flask

from abstra_internals.controllers.main import MainController


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("logs", __name__)

    @bp.get("/<id>")
    def _get_logs(id: str):
        logs = controller.get_logs(id)
        return [log.to_dto() for log in logs]

    return bp
